from __future__ import annotations

import copy
import logging
import time
from typing import Dict, FrozenSet, List, Optional, Sequence, Set, Tuple, Union

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.implementations.metric import PydanticMetricTimeWindow
from dbt_semantic_interfaces.protocols.metric import (
    ConstantPropertyInput,
    ConversionTypeParams,
    Metric,
    MetricInputMeasure,
    MetricTimeWindow,
    MetricType,
)
from dbt_semantic_interfaces.references import MeasureReference, MetricReference, TimeDimensionReference
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from dbt_semantic_interfaces.validations.unique_valid_name import MetricFlowReservedKeywords
from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId
from metricflow_semantics.errors.error_classes import UnableToSatisfyQueryError
from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.mf_logging.runtime import log_runtime
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.constant_property_spec import ConstantPropertySpec
from metricflow_semantics.specs.entity_spec import EntitySpec, LinklessEntitySpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet
from metricflow_semantics.specs.measure_spec import (
    CumulativeMeasureDescription,
    JoinToTimeSpineDescription,
    MeasureSpec,
    MetricInputMeasureSpec,
)
from metricflow_semantics.specs.metadata_spec import MetadataSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec
from metricflow_semantics.specs.order_by_spec import OrderBySpec
from metricflow_semantics.specs.query_spec import MetricFlowQuerySpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet, group_specs_by_type
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.specs.where_filter.where_filter_spec import WhereFilterSpec
from metricflow_semantics.specs.where_filter.where_filter_spec_set import WhereFilterSpecSet
from metricflow_semantics.specs.where_filter.where_filter_transform import WhereSpecFactory
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.time.dateutil_adjuster import DateutilTimePeriodAdjuster
from metricflow_semantics.time.granularity import ExpandedTimeGranularity
from metricflow_semantics.time.time_spine_source import TimeSpineSource

from metricflow.dataflow.builder.builder_cache import (
    BuildAnyMetricOutputNodeParameterSet,
    DataflowPlanBuilderCache,
    FindSourceNodeRecipeParameterSet,
    FindSourceNodeRecipeResult,
)
from metricflow.dataflow.builder.measure_spec_properties import MeasureSpecProperties
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.node_evaluator import (
    LinkableInstanceSatisfiabilityEvaluation,
    NodeEvaluatorForLinkableInstances,
)
from metricflow.dataflow.builder.source_node import SourceNodeBuilder, SourceNodeSet
from metricflow.dataflow.builder.source_node_recipe import SourceNodeRecipe
from metricflow.dataflow.dataflow_plan import (
    DataflowPlan,
    DataflowPlanNode,
)
from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
from metricflow.dataflow.nodes.aggregate_measures import AggregateMeasuresNode
from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode, SpecToAlias
from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinDescription, JoinOnEntitiesNode
from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.min_max import MinMaxNode
from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode
from metricflow.dataflow.optimizer.dataflow_optimizer_factory import (
    DataflowPlanOptimization,
    DataflowPlanOptimizerFactory,
)
from metricflow.dataset.dataset_classes import DataSet
from metricflow.plan_conversion.node_processor import (
    PredicateInputType,
    PredicatePushdownState,
    PreJoinNodeProcessor,
)

logger = logging.getLogger(__name__)


class DataflowPlanBuilder:
    """Builds a dataflow plan to satisfy a given query."""

    def __init__(  # noqa: D107
        self,
        source_node_set: SourceNodeSet,
        semantic_manifest_lookup: SemanticManifestLookup,
        node_output_resolver: DataflowPlanNodeOutputDataSetResolver,
        column_association_resolver: ColumnAssociationResolver,
        source_node_builder: SourceNodeBuilder,
        dataflow_plan_builder_cache: Optional[DataflowPlanBuilderCache] = None,
    ) -> None:
        self._semantic_model_lookup = semantic_manifest_lookup.semantic_model_lookup
        self._metric_lookup = semantic_manifest_lookup.metric_lookup
        self._metric_time_dimension_reference = DataSet.metric_time_dimension_reference()
        self._source_node_set = source_node_set
        self._column_association_resolver = column_association_resolver
        self._node_data_set_resolver = node_output_resolver
        self._source_node_builder = source_node_builder
        self._time_period_adjuster = DateutilTimePeriodAdjuster()
        self._cache = dataflow_plan_builder_cache or DataflowPlanBuilderCache()

    def build_plan(
        self,
        query_spec: MetricFlowQuerySpec,
        output_sql_table: Optional[SqlTable] = None,
        output_selection_specs: Optional[InstanceSpecSet] = None,
        optimizations: FrozenSet[DataflowPlanOptimization] = frozenset(),
    ) -> DataflowPlan:
        """Generate a plan for reading the results of a query with the given spec into a data_table or table."""
        # Workaround for a Pycharm type inspection issue with decorators.
        # noinspection PyArgumentList
        return self._build_plan(
            query_spec=query_spec,
            output_sql_table=output_sql_table,
            output_selection_specs=output_selection_specs,
            optimizations=optimizations,
        )

    def _build_query_output_node(
        self, query_spec: MetricFlowQuerySpec, for_group_by_source_node: bool = False
    ) -> DataflowPlanNode:
        """Build SQL output node from query inputs. May be used to build query DFP or source node."""
        metric_specs: Tuple[MetricSpec, ...] = ()
        for metric_spec in query_spec.metric_specs:
            if (
                len(metric_spec.filter_spec_set.all_filter_specs) > 0
                or metric_spec.offset_to_grain is not None
                or metric_spec.offset_window is not None
            ):
                # Remove aliases here. They will be added back at the very end of the query.
                metric_specs += (metric_spec.with_alias(None),) if metric_spec.alias else (metric_spec,)
                raise ValueError(
                    f"The metric specs in the query spec should not contain any metric modifiers. Got: {metric_spec}"
                )

        filter_spec_factory = WhereSpecFactory(
            column_association_resolver=self._column_association_resolver,
            spec_resolution_lookup=query_spec.filter_spec_resolution_lookup,
            semantic_model_lookup=self._semantic_model_lookup,
        )

        query_level_filter_specs = tuple(
            filter_spec_factory.create_from_where_filter_intersection(
                filter_location=WhereFilterLocation.for_query(
                    tuple(metric_spec.reference for metric_spec in query_spec.metric_specs)
                ),
                filter_intersection=query_spec.filter_intersection,
            )
        )

        predicate_pushdown_state = PredicatePushdownState(
            time_range_constraint=query_spec.time_range_constraint,
            where_filter_specs=(),
            pushdown_enabled_types=frozenset({PredicateInputType.TIME_RANGE_CONSTRAINT}),
        )
        return self._build_metrics_output_node(
            metric_specs=tuple(
                MetricSpec(
                    element_name=metric_spec.element_name,
                    filter_spec_set=WhereFilterSpecSet(query_level_filter_specs=query_level_filter_specs),
                )
                for metric_spec in query_spec.metric_specs
            ),
            queried_linkable_specs=query_spec.linkable_specs,
            filter_spec_factory=filter_spec_factory,
            predicate_pushdown_state=predicate_pushdown_state,
            for_group_by_source_node=for_group_by_source_node,
        )

    @log_runtime()
    def _build_plan(
        self,
        query_spec: MetricFlowQuerySpec,
        output_sql_table: Optional[SqlTable],
        output_selection_specs: Optional[InstanceSpecSet],
        optimizations: FrozenSet[DataflowPlanOptimization],
    ) -> DataflowPlan:
        metrics_output_node = self._build_query_output_node(query_spec=query_spec)

        sink_node = DataflowPlanBuilder.build_sink_node(
            parent_node=metrics_output_node,
            metric_specs=query_spec.metric_specs,
            order_by_specs=query_spec.order_by_specs,
            output_sql_table=output_sql_table,
            limit=query_spec.limit,
            output_selection_specs=output_selection_specs,
        )

        plan_id = DagId.from_id_prefix(StaticIdPrefix.DATAFLOW_PLAN_PREFIX)
        plan = DataflowPlan(sink_nodes=[sink_node], plan_id=plan_id)
        return self._optimize_plan(plan, optimizations)

    def _optimize_plan(self, plan: DataflowPlan, optimizations: FrozenSet[DataflowPlanOptimization]) -> DataflowPlan:
        optimizer_factory = DataflowPlanOptimizerFactory(self._node_data_set_resolver)
        for optimizer in optimizer_factory.get_optimizers(optimizations):
            logger.debug(LazyFormat(lambda: f"Applying optimizer: {optimizer.__class__.__name__}"))
            try:
                plan = optimizer.optimize(plan)
                logger.debug(
                    LazyFormat(
                        lambda: f"After applying optimizer {optimizer.__class__.__name__}, the dataflow plan is:\n"
                        f"{indent(plan.structure_text())}"
                    )
                )
            except Exception:
                logger.exception(f"Got an exception applying {optimizer.__class__.__name__}")

        return plan

    def _get_minimum_metric_time_spec_for_metric(self, metric_reference: MetricReference) -> TimeDimensionSpec:
        """Gets the minimum metric time spec for the given metric reference."""
        min_granularity = ExpandedTimeGranularity.from_time_granularity(
            self._metric_lookup.get_min_queryable_time_granularity(metric_reference)
        )
        return DataSet.metric_time_dimension_spec(min_granularity)

    def _build_aggregated_conversion_node(
        self,
        metric_spec: MetricSpec,
        base_measure_spec: MetricInputMeasureSpec,
        conversion_measure_spec: MetricInputMeasureSpec,
        entity_spec: EntitySpec,
        window: Optional[MetricTimeWindow],
        queried_linkable_specs: LinkableSpecSet,
        predicate_pushdown_state: PredicatePushdownState,
        constant_properties: Optional[Sequence[ConstantPropertyInput]] = None,
    ) -> DataflowPlanNode:
        """Builds a node that contains aggregated values of conversions and opportunities."""
        # Pushdown parameters vary with conversion metrics due to the way the time joins are applied.
        # Due to other outstanding issues with conversion metric filters, we disable predicate
        # pushdown for any filter parameter set that is not part of the original time range constraint
        # implementation.
        disabled_pushdown_state = PredicatePushdownState.with_pushdown_disabled()
        time_range_only_pushdown_state = PredicatePushdownState(
            time_range_constraint=predicate_pushdown_state.time_range_constraint,
            where_filter_specs=tuple(),
            pushdown_enabled_types=frozenset([PredicateInputType.TIME_RANGE_CONSTRAINT]),
        )

        # Build measure recipes
        base_required_linkable_specs = self.__get_required_linkable_specs(
            queried_linkable_specs=queried_linkable_specs,
            filter_specs=base_measure_spec.filter_spec_set.all_filter_specs,
        )
        base_measure_recipe = self._find_source_node_recipe(
            FindSourceNodeRecipeParameterSet(
                measure_spec_properties=self._build_measure_spec_properties([base_measure_spec.measure_spec]),
                predicate_pushdown_state=time_range_only_pushdown_state,
                linkable_spec_set=base_required_linkable_specs,
            )
        )
        logger.debug(LazyFormat(lambda: f"Recipe for base measure aggregation:\n{mf_pformat(base_measure_recipe)}"))
        conversion_measure_recipe = self._find_source_node_recipe(
            FindSourceNodeRecipeParameterSet(
                measure_spec_properties=self._build_measure_spec_properties([conversion_measure_spec.measure_spec]),
                predicate_pushdown_state=disabled_pushdown_state,
                linkable_spec_set=LinkableSpecSet(),
            )
        )
        logger.debug(
            LazyFormat(lambda: f"Recipe for conversion measure aggregation:\n{mf_pformat(conversion_measure_recipe)}")
        )
        if base_measure_recipe is None:
            raise UnableToSatisfyQueryError(
                f"Unable to join all items in request. Measure: {base_measure_spec.measure_spec}; Specs to join: {base_required_linkable_specs}"
            )
        if conversion_measure_recipe is None:
            raise UnableToSatisfyQueryError(
                f"Unable to build dataflow plan for conversion measure: {conversion_measure_spec.measure_spec}"
            )

        # Gets the aggregated opportunities
        aggregated_base_measure_node = self.build_aggregated_measure(
            metric_input_measure_spec=base_measure_spec,
            queried_linkable_specs=queried_linkable_specs,
            predicate_pushdown_state=time_range_only_pushdown_state,
        )

        # Build unaggregated conversions source node
        # Generate UUID column for conversion source to uniquely identify each row
        unaggregated_conversion_measure_node = AddGeneratedUuidColumnNode.create(
            parent_node=conversion_measure_recipe.source_node
        )

        # Get the time dimension used to calculate the conversion window
        # Currently, both the base/conversion measure uses metric_time as it's the default agg time dimension.
        # However, eventually, there can be user-specified time dimensions used for this calculation.
        min_metric_time_spec = self._get_minimum_metric_time_spec_for_metric(metric_spec.reference)

        # Filter the source nodes with only the required specs needed for the calculation
        constant_property_specs = []
        required_local_specs = [base_measure_spec.measure_spec, entity_spec, min_metric_time_spec] + list(
            base_measure_recipe.required_local_linkable_specs.as_tuple
        )
        for constant_property in constant_properties or []:
            base_property_spec = self._semantic_model_lookup.get_element_spec_for_name(constant_property.base_property)
            conversion_property_spec = self._semantic_model_lookup.get_element_spec_for_name(
                constant_property.conversion_property
            )
            required_local_specs.append(base_property_spec)
            constant_property_specs.append(
                ConstantPropertySpec(base_spec=base_property_spec, conversion_spec=conversion_property_spec)
            )

        # Build the unaggregated base measure node for computing conversions
        unaggregated_base_measure_node = self._build_pre_aggregation_plan(
            source_node=base_measure_recipe.source_node,
            join_targets=base_measure_recipe.join_targets,
            filter_to_specs=group_specs_by_type(required_local_specs)
            .merge(base_required_linkable_specs.as_instance_spec_set)
            .dedupe(),
            custom_granularity_specs=base_required_linkable_specs.time_dimension_specs_with_custom_grain,
            where_filter_specs=base_measure_spec.filter_spec_set.all_filter_specs,
        )

        # Gets the successful conversions using JoinConversionEventsNode
        # The conversion events are joined by the base events which are already time constrained. However, this could
        # be still be constrained, where we adjust the time range to the window size similar to cumulative, but
        # adjusted in the opposite direction.
        join_conversion_node = JoinConversionEventsNode.create(
            base_node=unaggregated_base_measure_node,
            base_time_dimension_spec=min_metric_time_spec,
            conversion_node=unaggregated_conversion_measure_node,
            conversion_measure_spec=conversion_measure_spec.measure_spec,
            conversion_time_dimension_spec=min_metric_time_spec,
            unique_identifier_keys=(MetadataSpec(MetricFlowReservedKeywords.MF_INTERNAL_UUID.value),),
            entity_spec=entity_spec,
            window=window,
            constant_properties=constant_property_specs,
        )

        # Aggregate the conversion events with the JoinConversionEventsNode as the source node.
        recipe_with_join_conversion_source_node = SourceNodeRecipe(
            source_node=join_conversion_node,
            required_local_linkable_specs=queried_linkable_specs,
            join_linkable_instances_recipes=(),
            all_linkable_specs_required_for_source_nodes=queried_linkable_specs,
        )
        # TODO: Refine conversion metric configuration to fit into the standard dataflow plan building model
        # In this case we override the measure recipe, which currently results in us bypassing predicate pushdown
        # Rather than relying on happenstance in the way the code is laid out we also explicitly disable
        # predicate pushdwon until we are ready to fully support it for conversion metrics
        aggregated_conversions_node = self.build_aggregated_measure(
            metric_input_measure_spec=conversion_measure_spec,
            queried_linkable_specs=queried_linkable_specs,
            predicate_pushdown_state=disabled_pushdown_state,
            measure_recipe=recipe_with_join_conversion_source_node,
        )

        # Combine the aggregated opportunities and conversion data sets
        return CombineAggregatedOutputsNode.create(
            parent_nodes=(aggregated_base_measure_node, aggregated_conversions_node)
        )

    def _build_conversion_metric_output_node(
        self,
        metric_spec: MetricSpec,
        queried_linkable_specs: LinkableSpecSet,
        filter_spec_factory: WhereSpecFactory,
        predicate_pushdown_state: PredicatePushdownState,
        for_group_by_source_node: bool = False,
    ) -> ComputeMetricsNode:
        """Builds a compute metric node for a conversion metric."""
        metric_reference = metric_spec.reference
        metric = self._metric_lookup.get_metric(metric_reference)
        conversion_type_params = metric.type_params.conversion_type_params
        assert conversion_type_params, "A conversion metric should have type_params.conversion_type_params defined."
        base_measure, conversion_measure = self._build_input_measure_specs_for_conversion_metric(
            metric_reference=metric_spec.reference,
            conversion_type_params=conversion_type_params,
            filter_spec_factory=filter_spec_factory,
            descendent_filter_spec_set=metric_spec.filter_spec_set,
            queried_linkable_specs=queried_linkable_specs,
        )
        # TODO: [custom granularity] change this to an assertion once we're sure there aren't exceptions
        if not StructuredLinkableSpecName.from_name(
            qualified_name=conversion_type_params.entity, custom_granularity_names=()
        ).is_element_name:
            logger.warning(
                LazyFormat(
                    lambda: f"Found additional annotations in type param entity name `{conversion_type_params.entity}`, which "
                    "should be a simple element name reference. This should have been blocked by model validation!"
                )
            )
        entity_spec = EntitySpec(element_name=conversion_type_params.entity, entity_links=())
        logger.debug(
            LazyFormat(
                lambda: f"For conversion metric {metric_spec},\n"
                f"base_measure is:\n{mf_pformat(base_measure)}\n"
                f"conversion_measure is:\n{mf_pformat(conversion_measure)}\n"
                f"entity is:\n{mf_pformat(entity_spec)}"
            )
        )

        aggregated_measures_node = self._build_aggregated_conversion_node(
            metric_spec=metric_spec,
            base_measure_spec=base_measure,
            conversion_measure_spec=conversion_measure,
            queried_linkable_specs=queried_linkable_specs,
            predicate_pushdown_state=predicate_pushdown_state,
            entity_spec=entity_spec,
            window=conversion_type_params.window,
            constant_properties=conversion_type_params.constant_properties,
        )

        return self.build_computed_metrics_node(
            metric_spec=metric_spec,
            aggregated_measures_node=aggregated_measures_node,
            for_group_by_source_node=for_group_by_source_node,
            aggregated_to_elements=set(queried_linkable_specs.as_tuple),
        )

    def _build_cumulative_metric_output_node(
        self,
        metric_spec: MetricSpec,
        queried_linkable_specs: LinkableSpecSet,
        filter_spec_factory: WhereSpecFactory,
        predicate_pushdown_state: PredicatePushdownState,
        for_group_by_source_node: bool = False,
    ) -> DataflowPlanNode:
        min_metric_time_spec = self._get_minimum_metric_time_spec_for_metric(metric_spec.reference)
        min_granularity = min_metric_time_spec.time_granularity

        queried_agg_time_dimensions = queried_linkable_specs.included_agg_time_dimension_specs_for_metric(
            metric_reference=metric_spec.reference, metric_lookup=self._metric_lookup
        )
        query_includes_agg_time_dimension_with_min_granularity = False
        for time_dimension_spec in queried_agg_time_dimensions:
            if time_dimension_spec.time_granularity == min_granularity:
                query_includes_agg_time_dimension_with_min_granularity = True
                break

        if query_includes_agg_time_dimension_with_min_granularity or len(queried_agg_time_dimensions) == 0:
            return self._build_base_metric_output_node(
                metric_spec=metric_spec,
                queried_linkable_specs=queried_linkable_specs,
                filter_spec_factory=filter_spec_factory,
                predicate_pushdown_state=predicate_pushdown_state,
                for_group_by_source_node=for_group_by_source_node,
            )

        # If a cumulative metric is queried without its minimum granularity, it will need to be aggregated twice:
        # once as a normal metric, and again using a window function to narrow down to one row per granularity period.
        # In this case, add metric time at the default granularity to the linkable specs. It will be used in the order by
        # clause of the window function and later excluded from the output selections.
        include_linkable_specs = queried_linkable_specs.add_specs(time_dimension_specs=(min_metric_time_spec,))
        compute_metrics_node = self._build_base_metric_output_node(
            metric_spec=metric_spec,
            queried_linkable_specs=include_linkable_specs,
            filter_spec_factory=filter_spec_factory,
            predicate_pushdown_state=predicate_pushdown_state,
            for_group_by_source_node=for_group_by_source_node,
        )
        return WindowReaggregationNode.create(
            parent_node=compute_metrics_node,
            metric_spec=metric_spec,
            order_by_spec=min_metric_time_spec,
            partition_by_specs=queried_linkable_specs.as_tuple,
        )

    def _build_base_metric_output_node(
        self,
        metric_spec: MetricSpec,
        queried_linkable_specs: LinkableSpecSet,
        filter_spec_factory: WhereSpecFactory,
        predicate_pushdown_state: PredicatePushdownState,
        for_group_by_source_node: bool = False,
    ) -> ComputeMetricsNode:
        """Builds a node to compute a metric that is not defined from other metrics."""
        metric_reference = metric_spec.reference
        metric = self._metric_lookup.get_metric(metric_reference)
        if metric.type is MetricType.SIMPLE or metric.type is MetricType.CUMULATIVE:
            pass
        elif (
            metric.type is MetricType.RATIO or metric.type is MetricType.DERIVED or metric.type is MetricType.CONVERSION
        ):
            raise ValueError(f"This should only be called for base metrics (simple or cumulative). Got: {metric.type}")
        else:
            assert_values_exhausted(metric.type)
        assert (
            len(metric.input_measures) == 1
        ), f"A base metric should not have multiple measures. Got {metric.input_measures}"

        metric_input_measure_spec = self._build_input_measure_spec(
            filter_spec_factory=filter_spec_factory,
            metric=metric,
            input_measure=metric.input_measures[0],
            queried_linkable_specs=queried_linkable_specs,
            child_metric_offset_window=metric_spec.offset_window,
            child_metric_offset_to_grain=metric_spec.offset_to_grain,
            cumulative_description=(
                CumulativeMeasureDescription(
                    cumulative_window=(
                        metric.type_params.cumulative_type_params.window
                        if metric.type_params.cumulative_type_params
                        else None
                    ),
                    cumulative_grain_to_date=(
                        metric.type_params.cumulative_type_params.grain_to_date
                        if metric.type_params.cumulative_type_params
                        else None
                    ),
                )
                if metric.type is MetricType.CUMULATIVE
                else None
            ),
            descendent_filter_spec_set=metric_spec.filter_spec_set,
        )
        logger.debug(
            LazyFormat(
                lambda: f"For\n{indent(mf_pformat(metric_spec))}"
                f"\nneeded measure is:"
                f"\n{indent(mf_pformat(metric_input_measure_spec))}"
            )
        )

        aggregated_measures_node = self.build_aggregated_measure(
            metric_input_measure_spec=metric_input_measure_spec,
            queried_linkable_specs=queried_linkable_specs,
            predicate_pushdown_state=predicate_pushdown_state,
        )
        return self.build_computed_metrics_node(
            metric_spec=metric_spec,
            aggregated_measures_node=aggregated_measures_node,
            for_group_by_source_node=for_group_by_source_node,
            aggregated_to_elements=set(queried_linkable_specs.as_tuple),
        )

    def _build_derived_metric_output_node(
        self,
        metric_spec: MetricSpec,
        queried_linkable_specs: LinkableSpecSet,
        filter_spec_factory: WhereSpecFactory,
        predicate_pushdown_state: PredicatePushdownState,
        for_group_by_source_node: bool = False,
    ) -> DataflowPlanNode:
        """Builds a node to compute a metric defined from other metrics."""
        metric = self._metric_lookup.get_metric(metric_spec.reference)
        metric_input_specs = self._build_input_metric_specs_for_derived_metric(
            metric_reference=metric_spec.reference,
            filter_spec_factory=filter_spec_factory,
        )
        logger.debug(
            LazyFormat(
                lambda: f"For {metric.type} metric: {metric_spec}, needed metrics are:\n"
                f"{mf_pformat(metric_input_specs)}"
            )
        )

        required_linkable_specs = self.__get_required_linkable_specs(
            queried_linkable_specs=queried_linkable_specs, filter_specs=metric_spec.filter_spec_set.all_filter_specs
        )

        parent_nodes: List[DataflowPlanNode] = []

        # This is the filter that's defined for the metric in the configs.
        metric_definition_filter_specs = filter_spec_factory.create_from_where_filter_intersection(
            filter_location=WhereFilterLocation.for_metric(metric_spec.reference),
            filter_intersection=metric.filter,
        )

        for metric_input_spec in metric_input_specs:
            where_filter_spec_set = WhereFilterSpecSet(
                metric_level_filter_specs=tuple(metric_definition_filter_specs),
            )

            # These are the filters that's defined as part of the input metric.
            where_filter_spec_set = where_filter_spec_set.merge(metric_input_spec.filter_spec_set)

            # If metric is offset, we'll apply where constraint after offset to avoid removing values
            # unexpectedly. Time constraint will be applied by INNER JOINing to time spine.
            # We may consider encapsulating this in pushdown state later, but as of this moment pushdown
            # is about post-join to pre-join for dimension access, and relies on the builder to collect
            # predicates from query and metric specs and make them available at measure level.
            if not metric_spec.has_time_offset:
                where_filter_spec_set = where_filter_spec_set.merge(metric_spec.filter_spec_set)
            metric_pushdown_state = (
                predicate_pushdown_state
                if not metric_spec.has_time_offset
                else PredicatePushdownState.with_pushdown_disabled()
            )

            parent_nodes.append(
                self._build_any_metric_output_node(
                    BuildAnyMetricOutputNodeParameterSet(
                        metric_spec=MetricSpec(
                            element_name=metric_input_spec.element_name,
                            filter_spec_set=where_filter_spec_set,
                            alias=metric_input_spec.alias,
                            offset_window=metric_input_spec.offset_window,
                            offset_to_grain=metric_input_spec.offset_to_grain,
                        ),
                        queried_linkable_specs=(
                            queried_linkable_specs if not metric_spec.has_time_offset else required_linkable_specs
                        ),
                        filter_spec_factory=filter_spec_factory,
                        predicate_pushdown_state=metric_pushdown_state,
                        for_group_by_source_node=False,
                    )
                )
            )

        parent_node = (
            parent_nodes[0]
            if len(parent_nodes) == 1
            else CombineAggregatedOutputsNode.create(parent_nodes=parent_nodes)
        )
        output_node: DataflowPlanNode = ComputeMetricsNode.create(
            parent_node=parent_node,
            metric_specs=[metric_spec],
            for_group_by_source_node=for_group_by_source_node,
            aggregated_to_elements=set(queried_linkable_specs.as_tuple),
        )

        # For ratio / derived metrics with time offset, apply offset join here. Constraints will be applied after the offset
        # to avoid filtering out values that will be changed.
        queried_agg_time_dimension_specs = queried_linkable_specs.included_agg_time_dimension_specs_for_metric(
            metric_reference=metric_spec.reference, metric_lookup=self._metric_lookup
        )
        if metric_spec.has_time_offset and queried_agg_time_dimension_specs:
            # TODO: move this to a helper method
            time_spine_node = self._build_time_spine_node(queried_agg_time_dimension_specs)
            output_node = JoinToTimeSpineNode.create(
                metric_source_node=output_node,
                time_spine_node=time_spine_node,
                requested_agg_time_dimension_specs=queried_agg_time_dimension_specs,
                join_on_time_dimension_spec=self._sort_by_base_granularity(queried_agg_time_dimension_specs)[0],
                offset_window=metric_spec.offset_window,
                offset_to_grain=metric_spec.offset_to_grain,
                join_type=SqlJoinType.INNER,
            )

            # TODO: fix bug here where filter specs are being included in when aggregating.
            if len(metric_spec.filter_spec_set.all_filter_specs) > 0 or predicate_pushdown_state.time_range_constraint:
                # FilterElementsNode will only be needed if there are where filter specs that were selected in the group by.
                specs_in_filters = set(
                    linkable_spec
                    for filter_spec in metric_spec.filter_spec_set.all_filter_specs
                    for linkable_spec in filter_spec.linkable_specs
                )
                filter_to_specs = None
                if not specs_in_filters.issubset(queried_linkable_specs.as_tuple):
                    filter_to_specs = InstanceSpecSet(metric_specs=(metric_spec,)).merge(
                        InstanceSpecSet.create_from_specs(queried_linkable_specs.as_tuple)
                    )
                output_node = self._build_pre_aggregation_plan(
                    source_node=output_node,
                    where_filter_specs=metric_spec.filter_spec_set.all_filter_specs,
                    time_range_constraint=predicate_pushdown_state.time_range_constraint,
                    filter_to_specs=filter_to_specs,
                )

        return output_node

    def _get_base_agg_time_dimensions(
        self, queried_linkable_specs: LinkableSpecSet, measure_reference: MeasureReference
    ) -> Tuple[TimeDimensionSpec, ...]:
        """Get queried agg_time_dimensios with their base grains, deduped.

        Custom grains are joined right before measure aggregation and after all other pre-aggregation joins,
        so only base grains are needed prior to that point.
        """
        queried_agg_time_dimension_specs = queried_linkable_specs.included_agg_time_dimension_specs_for_measure(
            measure_reference=measure_reference, semantic_model_lookup=self._semantic_model_lookup
        )
        return TimeDimensionSpec.with_base_grains(queried_agg_time_dimension_specs)

    def _build_any_metric_output_node(self, parameter_set: BuildAnyMetricOutputNodeParameterSet) -> DataflowPlanNode:
        """Builds a node to compute a metric of any type."""
        result = self._cache.get_build_any_metric_output_node_result(parameter_set)
        if result is not None:
            return result

        result = self._build_any_metric_output_node_non_cached(parameter_set)
        self._cache.set_build_any_metric_output_node_result(parameter_set, result)
        return result

    def _build_any_metric_output_node_non_cached(
        self, parameter_set: BuildAnyMetricOutputNodeParameterSet
    ) -> DataflowPlanNode:
        """Builds a node to compute a metric of any type."""
        metric_spec = parameter_set.metric_spec
        queried_linkable_specs = parameter_set.queried_linkable_specs
        filter_spec_factory = parameter_set.filter_spec_factory
        predicate_pushdown_state = parameter_set.predicate_pushdown_state
        for_group_by_source_node = parameter_set.for_group_by_source_node

        metric = self._metric_lookup.get_metric(metric_spec.reference)

        if metric.type is MetricType.SIMPLE:
            return self._build_base_metric_output_node(
                metric_spec=metric_spec,
                queried_linkable_specs=queried_linkable_specs,
                filter_spec_factory=filter_spec_factory,
                predicate_pushdown_state=predicate_pushdown_state,
                for_group_by_source_node=for_group_by_source_node,
            )

        elif metric.type is MetricType.CUMULATIVE:
            return self._build_cumulative_metric_output_node(
                metric_spec=metric_spec,
                queried_linkable_specs=queried_linkable_specs,
                filter_spec_factory=filter_spec_factory,
                predicate_pushdown_state=predicate_pushdown_state,
                for_group_by_source_node=for_group_by_source_node,
            )

        elif metric.type is MetricType.RATIO or metric.type is MetricType.DERIVED:
            return self._build_derived_metric_output_node(
                metric_spec=metric_spec,
                queried_linkable_specs=queried_linkable_specs,
                filter_spec_factory=filter_spec_factory,
                predicate_pushdown_state=predicate_pushdown_state,
                for_group_by_source_node=for_group_by_source_node,
            )
        elif metric.type is MetricType.CONVERSION:
            return self._build_conversion_metric_output_node(
                metric_spec=metric_spec,
                queried_linkable_specs=queried_linkable_specs,
                filter_spec_factory=filter_spec_factory,
                predicate_pushdown_state=predicate_pushdown_state,
                for_group_by_source_node=for_group_by_source_node,
            )

        assert_values_exhausted(metric.type)

    def _build_metrics_output_node(
        self,
        metric_specs: Sequence[MetricSpec],
        queried_linkable_specs: LinkableSpecSet,
        filter_spec_factory: WhereSpecFactory,
        predicate_pushdown_state: PredicatePushdownState,
        for_group_by_source_node: bool = False,
    ) -> DataflowPlanNode:
        """Builds a node that computes all requested metrics.

        Args:
            metric_specs: Specs for metrics to compute. Contains modifications to the metric defined in the model to
            include offsets and filters.
            queried_linkable_specs: Dimensions/entities that were queried.
            filter_spec_factory: Constructs WhereFilterSpecs with the resolved ambiguous group-by-items in the filter.
            predicate_pushdown_state: Parameters for evaluating and applying filter predicate pushdown, e.g., for
            applying time constraints prior to other dimension joins.
        """
        output_nodes: List[DataflowPlanNode] = []

        for metric_spec in metric_specs:
            logger.debug(LazyFormat(lambda: f"Generating compute metrics node for:\n{indent(mf_pformat(metric_spec))}"))
            self._metric_lookup.get_metric(metric_spec.reference)

            output_nodes.append(
                self._build_any_metric_output_node(
                    BuildAnyMetricOutputNodeParameterSet(
                        metric_spec=metric_spec,
                        queried_linkable_specs=queried_linkable_specs,
                        filter_spec_factory=filter_spec_factory,
                        predicate_pushdown_state=predicate_pushdown_state,
                        for_group_by_source_node=for_group_by_source_node,
                    )
                )
            )

        assert len(output_nodes) > 0, "ComputeMetricsNode was not properly constructed"

        if len(output_nodes) == 1:
            return output_nodes[0]

        return CombineAggregatedOutputsNode.create(parent_nodes=output_nodes)

    def build_plan_for_distinct_values(
        self, query_spec: MetricFlowQuerySpec, optimizations: FrozenSet[DataflowPlanOptimization] = frozenset()
    ) -> DataflowPlan:
        """Generate a plan that would get the distinct values of a linkable instance.

        e.g. distinct listing__country_latest for bookings by listing__country_latest
        """
        # Workaround for a Pycharm type inspection issue with decorators.
        # noinspection PyArgumentList
        return self._build_plan_for_distinct_values(query_spec, optimizations=optimizations)

    @log_runtime()
    def _build_plan_for_distinct_values(
        self, query_spec: MetricFlowQuerySpec, optimizations: FrozenSet[DataflowPlanOptimization]
    ) -> DataflowPlan:
        assert not query_spec.metric_specs, "Can't build distinct values plan with metrics."
        query_level_filter_specs: Sequence[WhereFilterSpec] = ()
        if query_spec.filter_intersection is not None and len(query_spec.filter_intersection.where_filters) > 0:
            filter_spec_factory = WhereSpecFactory(
                column_association_resolver=self._column_association_resolver,
                spec_resolution_lookup=query_spec.filter_spec_resolution_lookup
                or FilterSpecResolutionLookUp.empty_instance(),
                semantic_model_lookup=self._semantic_model_lookup,
            )

            query_level_filter_specs = filter_spec_factory.create_from_where_filter_intersection(
                filter_location=WhereFilterLocation.for_query(metric_references=tuple()),
                filter_intersection=query_spec.filter_intersection,
            )

        required_linkable_specs = self.__get_required_linkable_specs(
            queried_linkable_specs=query_spec.linkable_specs, filter_specs=query_level_filter_specs
        )
        predicate_pushdown_state = PredicatePushdownState(
            time_range_constraint=query_spec.time_range_constraint, where_filter_specs=tuple(query_level_filter_specs)
        )
        dataflow_recipe = self._find_source_node_recipe(
            FindSourceNodeRecipeParameterSet(
                linkable_spec_set=required_linkable_specs,
                predicate_pushdown_state=predicate_pushdown_state,
                measure_spec_properties=None,
            )
        )
        if not dataflow_recipe:
            raise UnableToSatisfyQueryError(f"Unable to join all items in request: {required_linkable_specs}")

        output_node = self._build_pre_aggregation_plan(
            source_node=dataflow_recipe.source_node,
            join_targets=dataflow_recipe.join_targets,
            filter_to_specs=InstanceSpecSet.create_from_specs(query_spec.linkable_specs.as_tuple),
            custom_granularity_specs=required_linkable_specs.time_dimension_specs_with_custom_grain,
            where_filter_specs=query_level_filter_specs,
            time_range_constraint=query_spec.time_range_constraint,
            distinct=True,
        )

        if query_spec.min_max_only:
            output_node = MinMaxNode.create(parent_node=output_node)

        sink_node = self.build_sink_node(
            parent_node=output_node,
            metric_specs=query_spec.metric_specs,
            order_by_specs=query_spec.order_by_specs,
            limit=query_spec.limit,
        )

        plan = DataflowPlan(sink_nodes=[sink_node])
        return self._optimize_plan(plan, optimizations)

    @staticmethod
    def build_sink_node(
        parent_node: DataflowPlanNode,
        metric_specs: Sequence[MetricSpec],
        order_by_specs: Sequence[OrderBySpec],
        output_sql_table: Optional[SqlTable] = None,
        limit: Optional[int] = None,
        output_selection_specs: Optional[InstanceSpecSet] = None,
    ) -> DataflowPlanNode:
        """Adds order by / limit / write nodes."""
        pre_result_node: Optional[DataflowPlanNode] = None

        if order_by_specs or limit:
            pre_result_node = OrderByLimitNode.create(
                order_by_specs=list(order_by_specs), limit=limit, parent_node=parent_node
            )

        if output_selection_specs:
            pre_result_node = FilterElementsNode.create(
                parent_node=pre_result_node or parent_node, include_specs=output_selection_specs
            )

        alias_specs = tuple(
            SpecToAlias(MetricSpec(metric.element_name), MetricSpec(metric.element_name, alias=metric.alias))
            for metric in metric_specs
            if metric.alias is not None
        )
        if len(alias_specs) > 0:
            pre_result_node = AliasSpecsNode.create(
                parent_node=pre_result_node or parent_node, change_specs=alias_specs
            )

        write_result_node: DataflowPlanNode
        if not output_sql_table:
            write_result_node = WriteToResultDataTableNode.create(parent_node=pre_result_node or parent_node)
        else:
            write_result_node = WriteToResultTableNode.create(
                parent_node=pre_result_node or parent_node, output_sql_table=output_sql_table
            )

        return write_result_node

    @staticmethod
    def _contains_multihop_linkables(linkable_specs: Sequence[LinkableInstanceSpec]) -> bool:
        """Returns true if any of the linkable specs requires a multi-hop join to realize."""
        return any(len(x.entity_links) > 1 for x in linkable_specs)

    def _sort_by_suitability(self, nodes: Sequence[DataflowPlanNode]) -> Sequence[DataflowPlanNode]:
        """Sort nodes by the number of linkable specs.

        The lower the number of linkable specs means less aggregation required.
        """

        def sort_function(node: DataflowPlanNode) -> int:
            data_set = self._node_data_set_resolver.get_output_data_set(node)
            return len(data_set.instance_set.spec_set.linkable_specs)

        return sorted(nodes, key=sort_function)

    def _select_source_nodes_with_measures(
        self, measure_specs: Set[MeasureSpec], source_nodes: Sequence[DataflowPlanNode]
    ) -> Sequence[DataflowPlanNode]:
        nodes = []
        measure_specs_set = set(measure_specs)
        for source_node in source_nodes:
            measure_specs_in_node = self._node_data_set_resolver.get_output_data_set(
                source_node
            ).instance_set.spec_set.measure_specs
            if measure_specs_set.intersection(set(measure_specs_in_node)) == measure_specs_set:
                nodes.append(source_node)
        return nodes

    def _select_source_nodes_with_linkable_specs(
        self, linkable_specs: LinkableSpecSet, source_nodes: Sequence[DataflowPlanNode]
    ) -> Sequence[DataflowPlanNode]:
        """Find source nodes with requested linkable specs and no measures."""
        # Use a dictionary to dedupe for consistent ordering.
        selected_nodes: Dict[DataflowPlanNode, None] = {}

        # TODO: Add support for no-metrics queries for custom grains without a join (i.e., select directly from time spine).
        linkable_specs_set = set(linkable_specs.as_tuple)
        for source_node in source_nodes:
            output_spec_set = self._node_data_set_resolver.get_output_data_set(source_node).instance_set.spec_set
            all_linkable_specs_in_node = set(output_spec_set.linkable_specs)
            requested_linkable_specs_in_node = linkable_specs_set.intersection(all_linkable_specs_in_node)
            if requested_linkable_specs_in_node:
                selected_nodes[source_node] = None

        return tuple(selected_nodes.keys())

    def _find_non_additive_dimension_in_linkable_specs(
        self,
        agg_time_dimension: TimeDimensionReference,
        linkable_specs: Sequence[LinkableInstanceSpec],
        non_additive_dimension_spec: NonAdditiveDimensionSpec,
    ) -> Optional[TimeDimensionSpec]:
        """Finds the TimeDimensionSpec matching the non_additive_dimension_spec, if any."""
        queried_time_dimension_spec: Optional[LinkableInstanceSpec] = None
        for linkable_spec in linkable_specs:
            dimension_name_match = linkable_spec.element_name == non_additive_dimension_spec.name
            metric_time_match = (
                non_additive_dimension_spec.name == agg_time_dimension.element_name
                and linkable_spec.element_name == self._metric_time_dimension_reference.element_name
            )
            if dimension_name_match or metric_time_match:
                queried_time_dimension_spec = linkable_spec
                break
        assert queried_time_dimension_spec is None or isinstance(
            queried_time_dimension_spec, TimeDimensionSpec
        ), "Non-additive dimension can only be a time dimension, if specified."
        return queried_time_dimension_spec

    def _build_measure_spec_properties(self, measure_specs: Sequence[MeasureSpec]) -> MeasureSpecProperties:
        """Ensures that the group of MeasureSpecs has the same non_additive_dimension_spec and agg_time_dimension."""
        if len(measure_specs) == 0:
            raise ValueError("Cannot build MeasureParametersForRecipe when given an empty sequence of measure_specs.")
        semantic_model_names = {
            self._semantic_model_lookup.measure_lookup.get_properties(
                measure.reference
            ).model_reference.semantic_model_name
            for measure in measure_specs
        }
        if len(semantic_model_names) > 1:
            raise ValueError(
                f"Cannot find common properties for measures {measure_specs} coming from multiple "
                f"semantic models: {semantic_model_names}. This suggests the measure_specs were not correctly filtered."
            )
        semantic_model_name = semantic_model_names.pop()

        measure_reference = measure_specs[0].reference
        measure_properties = self._semantic_model_lookup.measure_lookup.get_properties(measure_reference)
        agg_time_dimension = measure_properties.agg_time_dimension_reference
        agg_time_dimension_grain = measure_properties.agg_time_granularity

        non_additive_dimension_spec = measure_specs[0].non_additive_dimension_spec
        for measure_spec in measure_specs:
            if non_additive_dimension_spec != measure_spec.non_additive_dimension_spec:
                raise ValueError(f"measure_specs {measure_specs} do not have the same non_additive_dimension_spec.")
            measure_agg_time_dimension = self._semantic_model_lookup.measure_lookup.get_properties(
                measure_spec.reference
            ).agg_time_dimension_reference
            if measure_agg_time_dimension != agg_time_dimension:
                raise ValueError(f"measure_specs {measure_specs} do not have the same agg_time_dimension.")
        return MeasureSpecProperties(
            measure_specs=tuple(measure_specs),
            semantic_model_name=semantic_model_name,
            agg_time_dimension=agg_time_dimension,
            agg_time_dimension_grain=agg_time_dimension_grain,
            non_additive_dimension_spec=non_additive_dimension_spec,
        )

    def _find_source_node_recipe(self, parameter_set: FindSourceNodeRecipeParameterSet) -> Optional[SourceNodeRecipe]:
        """Find the most suitable source nodes to satisfy the requested specs, as well as how to join them."""
        result = self._cache.get_find_source_node_recipe_result(parameter_set)
        if result is not None:
            return result.source_node_recipe
        source_node_recipe = self._find_source_node_recipe_non_cached(parameter_set)
        self._cache.set_find_source_node_recipe_result(parameter_set, FindSourceNodeRecipeResult(source_node_recipe))
        return source_node_recipe

    def _find_source_node_recipe_non_cached(
        self, parameter_set: FindSourceNodeRecipeParameterSet
    ) -> Optional[SourceNodeRecipe]:
        linkable_spec_set = parameter_set.linkable_spec_set
        predicate_pushdown_state = parameter_set.predicate_pushdown_state
        measure_spec_properties = parameter_set.measure_spec_properties

        candidate_nodes_for_left_side_of_join: List[DataflowPlanNode] = []
        candidate_nodes_for_right_side_of_join: List[DataflowPlanNode] = []

        # Replace any custom granularities with their base granularities. The custom granularity will be joined in
        # later, since custom granularities cannot be satisfied by source nodes. But we will need the dimension at
        # base granularity from the source node in order to join to the appropriate time spine later.
        linkable_specs_to_satisfy = linkable_spec_set.replace_custom_granularity_with_base_granularity()
        linkable_specs_to_satisfy_tuple = linkable_specs_to_satisfy.as_tuple
        if measure_spec_properties:
            candidate_nodes_for_right_side_of_join += self._source_node_set.source_nodes_for_metric_queries
            candidate_nodes_for_left_side_of_join += self._select_source_nodes_with_measures(
                measure_specs=set(measure_spec_properties.measure_specs),
                source_nodes=self._source_node_set.source_nodes_for_metric_queries,
            )
            default_join_type = SqlJoinType.LEFT_OUTER
        else:
            candidate_nodes_for_right_side_of_join += list(self._source_node_set.source_nodes_for_group_by_item_queries)
            candidate_nodes_for_left_side_of_join += list(
                self._select_source_nodes_with_linkable_specs(
                    linkable_specs=linkable_specs_to_satisfy,
                    source_nodes=self._source_node_set.source_nodes_for_group_by_item_queries,
                )
            )
            # If metric_time is requested without metrics, choose appropriate time spine node to select those values from.
            if linkable_specs_to_satisfy.metric_time_specs:
                time_spine_node = self._choose_time_spine_metric_time_node(linkable_specs_to_satisfy.metric_time_specs)
                candidate_nodes_for_right_side_of_join += [time_spine_node]
                candidate_nodes_for_left_side_of_join += [time_spine_node]
            default_join_type = SqlJoinType.FULL_OUTER

        logger.debug(
            LazyFormat(
                lambda: f"Starting search with {len(candidate_nodes_for_left_side_of_join)} potential source nodes on the left "
                f"side of the join, and {len(candidate_nodes_for_right_side_of_join)} potential nodes on the right side "
                f"of the join."
            )
        )
        start_time = time.time()

        node_processor = PreJoinNodeProcessor(
            semantic_model_lookup=self._semantic_model_lookup,
            node_data_set_resolver=self._node_data_set_resolver,
        )

        if predicate_pushdown_state.has_pushdown_potential and default_join_type is not SqlJoinType.FULL_OUTER:
            # TODO: encapsulate join type and distinct values state and eventually move this to a DataflowPlanOptimizer
            # This works today because all of our subsequent join configuration operations preserve the join type
            # as-is, or else switch it to a CROSS JOIN or INNER JOIN type, both of which are safe for predicate
            # pushdown. However, there is currently no way to enforce that invariant, so we will need to move
            # to a model where we evaluate the join nodes themselves and decide on whether or not to push down
            # the predicate. This will be much more straightforward once we finish encapsulating our existing
            # time range constraint pushdown controls into this mechanism.
            candidate_nodes_for_left_side_of_join = list(
                node_processor.apply_matching_filter_predicates(
                    source_nodes=candidate_nodes_for_left_side_of_join,
                    predicate_pushdown_state=predicate_pushdown_state,
                    metric_time_dimension_reference=self._metric_time_dimension_reference,
                )
            )

        candidate_nodes_for_right_side_of_join = node_processor.remove_unnecessary_nodes(
            desired_linkable_specs=linkable_specs_to_satisfy_tuple,
            nodes=candidate_nodes_for_right_side_of_join,
            metric_time_dimension_reference=self._metric_time_dimension_reference,
            time_spine_metric_time_nodes=self._source_node_set.time_spine_metric_time_nodes_tuple,
        )
        logger.debug(
            LazyFormat(
                lambda: f"After removing unnecessary nodes, there are {len(candidate_nodes_for_right_side_of_join)} candidate "
                f"nodes for the right side of the join"
            )
        )
        # TODO: test multi-hop with custom grains
        if DataflowPlanBuilder._contains_multihop_linkables(linkable_specs_to_satisfy_tuple):
            candidate_nodes_for_right_side_of_join = list(
                node_processor.add_multi_hop_joins(
                    desired_linkable_specs=linkable_specs_to_satisfy_tuple,
                    nodes=candidate_nodes_for_right_side_of_join,
                    join_type=default_join_type,
                )
            )
            logger.debug(
                LazyFormat(
                    lambda: f"After adding multi-hop nodes, there are {len(candidate_nodes_for_right_side_of_join)} candidate "
                    f"nodes for the right side of the join:\n"
                    f"{mf_pformat(candidate_nodes_for_right_side_of_join)}"
                )
            )

        # If there are MetricGroupBys in the requested linkable specs, build source nodes to satisfy them.
        # We do this at query time instead of during usual source node generation because the number of potential
        # MetricGroupBy source nodes could be extremely large (and potentially slow).
        logger.debug(
            LazyFormat(
                lambda: f"Building source nodes for group by metrics: {linkable_specs_to_satisfy.group_by_metric_specs}"
            )
        )
        candidate_nodes_for_right_side_of_join += [
            self._build_query_output_node(
                query_spec=self._source_node_builder.build_source_node_inputs_for_group_by_metric(group_by_metric_spec),
                for_group_by_source_node=True,
            )
            for group_by_metric_spec in linkable_specs_to_satisfy.group_by_metric_specs
        ]

        logger.debug(LazyFormat(lambda: f"Processing nodes took: {time.time()-start_time:.2f}s"))

        node_evaluator = NodeEvaluatorForLinkableInstances(
            semantic_model_lookup=self._semantic_model_lookup,
            nodes_available_for_joins=self._sort_by_suitability(candidate_nodes_for_right_side_of_join),
            node_data_set_resolver=self._node_data_set_resolver,
            time_spine_metric_time_nodes=self._source_node_set.time_spine_metric_time_nodes_tuple,
        )

        # Dict from the node that contains the source node to the evaluation results.
        node_to_evaluation: Dict[DataflowPlanNode, LinkableInstanceSatisfiabilityEvaluation] = {}

        for node in self._sort_by_suitability(candidate_nodes_for_left_side_of_join):
            data_set = self._node_data_set_resolver.get_output_data_set(node)

            if measure_spec_properties:
                measure_specs = measure_spec_properties.measure_specs
                missing_specs = [
                    spec for spec in measure_specs if spec not in data_set.instance_set.spec_set.measure_specs
                ]
                if missing_specs:
                    logger.debug(
                        LazyFormat(
                            lambda: f"Skipping evaluation for:\n"
                            f"{indent(node.structure_text())}"
                            f"since it does not have all of the measure specs:\n"
                            f"{indent(mf_pformat(missing_specs))}"
                        )
                    )
                    continue

            logger.debug(
                LazyFormat(
                    lambda: f"Evaluating candidate node for the left side of the join:\n{indent(mf_pformat(node.structure_text()))}"
                )
            )

            start_time = time.time()
            evaluation = node_evaluator.evaluate_node(
                left_node=node,
                required_linkable_specs=list(linkable_specs_to_satisfy_tuple),
                default_join_type=default_join_type,
            )
            logger.debug(LazyFormat(lambda: f"Evaluation of {node} took {time.time() - start_time:.2f}s"))

            logger.debug(
                LazyFormat(
                    lambda: "Evaluation for source node:"
                    + indent(f"\nnode:\n{indent(node.structure_text())}")
                    + indent(f"\nevaluation:\n{indent(mf_pformat(evaluation))}")
                )
            )

            if len(evaluation.unjoinable_linkable_specs) > 0:
                logger.debug(
                    LazyFormat(
                        lambda: f"Skipping {node.node_id} since it contains un-joinable specs: "
                        f"{evaluation.unjoinable_linkable_specs}"
                    )
                )
                continue

            num_joins_required = len(evaluation.join_recipes)
            logger.debug(
                LazyFormat(
                    lambda: f"Found candidate with node ID '{node.node_id}' with {num_joins_required} joins required."
                )
            )

            node_to_evaluation[node] = evaluation

            # Since are evaluating nodes with the lowest cost first, if we find one without requiring any joins, then
            # this is going to be the lowest cost solution.
            if len(evaluation.join_recipes) == 0:
                logger.debug(
                    LazyFormat(lambda: "Not evaluating other nodes since we found one that doesn't require joins")
                )
                break

        logger.debug(LazyFormat(lambda: f"Found {len(node_to_evaluation)} candidate source nodes."))

        if len(node_to_evaluation) > 0:
            # Find evaluation with lowest number of joins.
            node_with_lowest_cost_plan = min(
                node_to_evaluation, key=lambda node: len(node_to_evaluation[node].join_recipes)
            )
            evaluation = node_to_evaluation[node_with_lowest_cost_plan]

            logger.debug(
                LazyFormat(
                    lambda: "Lowest cost plan is:"
                    + indent(f"\nnode:\n{indent(node_with_lowest_cost_plan.structure_text())}")
                    + indent(f"\nevaluation:\n{indent(mf_pformat(evaluation))}")
                    + indent(f"\njoins: {len(node_to_evaluation[node_with_lowest_cost_plan].join_recipes)}")
                )
            )

            # Nodes containing the linkable instances will be joined to the source node, so these
            # entities will need to be present in the source node.
            required_local_entity_specs = tuple(x.join_on_entity for x in evaluation.join_recipes if x.join_on_entity)
            # Same thing with partitions.
            required_local_dimension_specs = tuple(
                y.start_node_dimension_spec for x in evaluation.join_recipes for y in x.join_on_partition_dimensions
            )
            required_local_time_dimension_specs = tuple(
                y.start_node_time_dimension_spec
                for x in evaluation.join_recipes
                for y in x.join_on_partition_time_dimensions
            )
            return SourceNodeRecipe(
                source_node=node_with_lowest_cost_plan,
                required_local_linkable_specs=LinkableSpecSet.create_from_specs(
                    evaluation.local_linkable_specs
                    + required_local_entity_specs
                    + required_local_dimension_specs
                    + required_local_time_dimension_specs
                ),
                join_linkable_instances_recipes=node_to_evaluation[node_with_lowest_cost_plan].join_recipes,
                all_linkable_specs_required_for_source_nodes=linkable_specs_to_satisfy,
            )

        logger.error(LazyFormat(lambda: "No recipe could be constructed."))
        return None

    def build_computed_metrics_node(
        self,
        metric_spec: MetricSpec,
        aggregated_measures_node: Union[AggregateMeasuresNode, DataflowPlanNode],
        aggregated_to_elements: Set[LinkableInstanceSpec],
        for_group_by_source_node: bool = False,
    ) -> ComputeMetricsNode:
        """Builds a ComputeMetricsNode from aggregated measures."""
        return ComputeMetricsNode.create(
            parent_node=aggregated_measures_node,
            metric_specs=[metric_spec],
            for_group_by_source_node=for_group_by_source_node,
            aggregated_to_elements=aggregated_to_elements,
        )

    def _build_input_measure_specs_for_conversion_metric(
        self,
        metric_reference: MetricReference,
        conversion_type_params: ConversionTypeParams,
        filter_spec_factory: WhereSpecFactory,
        descendent_filter_spec_set: WhereFilterSpecSet,
        queried_linkable_specs: LinkableSpecSet,
    ) -> Tuple[MetricInputMeasureSpec, MetricInputMeasureSpec]:
        """Return [base_measure_input, conversion_measure_input] for computing a conversion metric."""
        metric = self._metric_lookup.get_metric(metric_reference)
        if metric.type is not MetricType.CONVERSION:
            raise ValueError("This should only be called for conversion metrics.")

        base_input_measure, conversion_input_measure = [
            self._build_input_measure_spec(
                filter_spec_factory=filter_spec_factory,
                metric=metric,
                input_measure=input_measure,
                queried_linkable_specs=queried_linkable_specs,
                descendent_filter_spec_set=descendent_filter_spec_set,
                include_filters=include_filters,
            )
            # Filters should only be applied to base measures.
            for input_measure, include_filters in [
                (conversion_type_params.base_measure, True),
                (conversion_type_params.conversion_measure, False),
            ]
        ]

        return base_input_measure, conversion_input_measure

    def _build_input_measure_spec(
        self,
        filter_spec_factory: WhereSpecFactory,
        metric: Metric,
        input_measure: MetricInputMeasure,
        descendent_filter_spec_set: WhereFilterSpecSet,
        queried_linkable_specs: LinkableSpecSet,
        include_filters: bool = True,
        child_metric_offset_window: Optional[MetricTimeWindow] = None,
        child_metric_offset_to_grain: Optional[TimeGranularity] = None,
        cumulative_description: Optional[CumulativeMeasureDescription] = None,
    ) -> MetricInputMeasureSpec:
        """Return the input measure spec required to compute the base metric.

        "child" refers to the derived metric that uses the metric specified by metric_reference in the definition.
        descendent_filter_specs includes all filter specs required to compute the metric in the query. This includes the
        filters in the query and any filter in the definition of metrics in between.
        """
        filter_spec_set: WhereFilterSpecSet = WhereFilterSpecSet()
        if include_filters:
            filter_spec_set = self._build_filter_specs_for_input_measure(
                filter_spec_factory=filter_spec_factory,
                metric=metric,
                input_measure=input_measure,
                descendent_filter_spec_set=descendent_filter_spec_set,
            )

        measure_spec = MeasureSpec(
            element_name=input_measure.name,
            non_additive_dimension_spec=self._semantic_model_lookup.non_additive_dimension_specs_by_measure.get(
                input_measure.measure_reference
            ),
        )

        before_aggregation_time_spine_join_description = None
        # If querying an offset metric, join to time spine.
        if child_metric_offset_window is not None or child_metric_offset_to_grain is not None:
            before_aggregation_time_spine_join_description = JoinToTimeSpineDescription(
                join_type=SqlJoinType.INNER,
                offset_window=child_metric_offset_window,
                offset_to_grain=child_metric_offset_to_grain,
            )

        # Even if the measure is configured to join to time spine, if there's no agg_time_dimension in the query,
        # there's no need to join to the time spine since all time will be aggregated.
        after_aggregation_time_spine_join_description = None
        if input_measure.join_to_timespine:
            if (
                len(
                    queried_linkable_specs.included_agg_time_dimension_specs_for_measure(
                        measure_reference=measure_spec.reference, semantic_model_lookup=self._semantic_model_lookup
                    )
                )
                > 0
            ):
                after_aggregation_time_spine_join_description = JoinToTimeSpineDescription(
                    join_type=SqlJoinType.LEFT_OUTER,
                    offset_window=None,
                    offset_to_grain=None,
                )

        return MetricInputMeasureSpec(
            measure_spec=measure_spec,
            fill_nulls_with=input_measure.fill_nulls_with,
            offset_window=child_metric_offset_window,
            offset_to_grain=child_metric_offset_to_grain,
            cumulative_description=cumulative_description,
            filter_spec_set=filter_spec_set,
            alias=input_measure.alias,
            before_aggregation_time_spine_join_description=before_aggregation_time_spine_join_description,
            after_aggregation_time_spine_join_description=after_aggregation_time_spine_join_description,
        )

    def _build_filter_specs_for_input_measure(
        self,
        filter_spec_factory: WhereSpecFactory,
        metric: Metric,
        input_measure: MetricInputMeasure,
        descendent_filter_spec_set: WhereFilterSpecSet,
    ) -> WhereFilterSpecSet:
        metric_reference = MetricReference(element_name=metric.name)
        filter_location = WhereFilterLocation.for_metric(metric_reference)

        filter_spec_set = WhereFilterSpecSet(
            measure_level_filter_specs=tuple(
                filter_spec_factory.create_from_where_filter_intersection(
                    filter_location=filter_location, filter_intersection=input_measure.filter
                )
            ),
            metric_level_filter_specs=tuple(
                filter_spec_factory.create_from_where_filter_intersection(
                    filter_location=filter_location, filter_intersection=metric.filter
                )
            ),
        )
        return filter_spec_set.merge(descendent_filter_spec_set)

    def _build_input_metric_specs_for_derived_metric(
        self,
        metric_reference: MetricReference,
        filter_spec_factory: WhereSpecFactory,
    ) -> Sequence[MetricSpec]:
        """Return the metric specs referenced by the metric. Current use case is for derived metrics."""
        metric = self._metric_lookup.get_metric(metric_reference)
        input_metric_specs: List[MetricSpec] = []

        for input_metric in metric.input_metrics:
            filter_spec_set = WhereFilterSpecSet(
                metric_level_filter_specs=tuple(
                    filter_spec_factory.create_from_where_filter_intersection(
                        filter_location=WhereFilterLocation.for_input_metric(
                            input_metric_reference=input_metric.as_reference
                        ),
                        filter_intersection=input_metric.filter,
                    )
                ),
            )

            spec = MetricSpec(
                element_name=input_metric.name,
                filter_spec_set=filter_spec_set,
                alias=input_metric.alias,
                offset_window=(
                    PydanticMetricTimeWindow(
                        count=input_metric.offset_window.count,
                        granularity=input_metric.offset_window.granularity,
                    )
                    if input_metric.offset_window
                    else None
                ),
                offset_to_grain=input_metric.offset_to_grain,
            )
            input_metric_specs.append(spec)
        return tuple(input_metric_specs)

    def build_aggregated_measure(
        self,
        metric_input_measure_spec: MetricInputMeasureSpec,
        queried_linkable_specs: LinkableSpecSet,
        predicate_pushdown_state: PredicatePushdownState,
        measure_recipe: Optional[SourceNodeRecipe] = None,
    ) -> DataflowPlanNode:
        """Returns a node where the measures are aggregated by the linkable specs and constrained appropriately.

        This might be a node representing a single aggregation over one semantic model, or a node representing
        a composite set of aggregations originating from multiple semantic models, and joined into a single
        aggregated set of measures.
        """
        measure_spec = metric_input_measure_spec.measure_spec

        logger.debug(
            LazyFormat(
                lambda: f"Building aggregated measure: {measure_spec} with input measure filters:\n"
                f"{mf_pformat(metric_input_measure_spec.filter_spec_set)}\n"
                f"and  filters:\n{mf_pformat(metric_input_measure_spec.filter_spec_set)}"
            )
        )

        return self._build_aggregated_measure_from_measure_source_node(
            metric_input_measure_spec=metric_input_measure_spec,
            queried_linkable_specs=queried_linkable_specs,
            predicate_pushdown_state=predicate_pushdown_state,
            measure_recipe=measure_recipe,
        )

    def __get_required_linkable_specs(
        self,
        queried_linkable_specs: LinkableSpecSet,
        filter_specs: Sequence[WhereFilterSpec],
        measure_spec_properties: Optional[MeasureSpecProperties] = None,
    ) -> LinkableSpecSet:
        """Get all required linkable specs for this query, including extraneous linkable specs.

        Extraneous linkable specs are specs that are used in this phase that should not show up in the final result
        unless it was already a requested spec in the query, e.g., a linkable spec used in where constraint is extraneous.
        """
        linkable_spec_sets_to_merge: List[LinkableSpecSet] = []
        for filter_spec in filter_specs:
            linkable_spec_sets_to_merge.append(LinkableSpecSet.create_from_specs(filter_spec.linkable_specs))

        if measure_spec_properties is not None:
            non_additive_dimension_spec = (
                measure_spec_properties.non_additive_dimension_spec if measure_spec_properties else None
            )
            if non_additive_dimension_spec is not None:
                agg_time_dimension_grain = measure_spec_properties.agg_time_dimension_grain
                linkable_spec_sets_to_merge.append(
                    LinkableSpecSet.create_from_specs(
                        non_additive_dimension_spec.linkable_specs(agg_time_dimension_grain)
                    )
                )

        extraneous_linkable_specs = LinkableSpecSet.merge_iterable(linkable_spec_sets_to_merge).dedupe()
        required_linkable_specs = queried_linkable_specs.merge(extraneous_linkable_specs).dedupe()

        # Custom grains require joining to their base grain, so add base grain to extraneous specs.
        if required_linkable_specs.time_dimension_specs_with_custom_grain:
            base_grain_set = LinkableSpecSet.create_from_specs(
                [spec.with_base_grain() for spec in required_linkable_specs.time_dimension_specs_with_custom_grain]
            )
            extraneous_linkable_specs = extraneous_linkable_specs.merge(base_grain_set).dedupe()
            required_linkable_specs = required_linkable_specs.merge(extraneous_linkable_specs).dedupe()

        return required_linkable_specs

    def _build_aggregated_measure_from_measure_source_node(
        self,
        metric_input_measure_spec: MetricInputMeasureSpec,
        queried_linkable_specs: LinkableSpecSet,
        predicate_pushdown_state: PredicatePushdownState,
        measure_recipe: Optional[SourceNodeRecipe] = None,
    ) -> DataflowPlanNode:
        measure_spec = metric_input_measure_spec.measure_spec
        cumulative = metric_input_measure_spec.cumulative_description is not None
        cumulative_window = (
            metric_input_measure_spec.cumulative_description.cumulative_window
            if metric_input_measure_spec.cumulative_description is not None
            else None
        )
        cumulative_grain_to_date = (
            metric_input_measure_spec.cumulative_description.cumulative_grain_to_date
            if metric_input_measure_spec.cumulative_description
            else None
        )
        measure_properties = self._build_measure_spec_properties([measure_spec])
        cumulative_metric_adjusted_time_constraint: Optional[TimeRangeConstraint] = None
        if cumulative and predicate_pushdown_state.time_range_constraint is not None:
            logger.debug(
                LazyFormat(
                    lambda: f"Time range constraint before adjustment is {predicate_pushdown_state.time_range_constraint}"
                )
            )
            granularity: Optional[TimeGranularity] = None
            count = 0
            if cumulative_window is not None:
                granularity = cumulative_window.granularity
                count = cumulative_window.count
            elif cumulative_grain_to_date is not None:
                count = 1
                granularity = cumulative_grain_to_date

            cumulative_metric_adjusted_time_constraint = (
                self._time_period_adjuster.expand_time_constraint_for_cumulative_metric(
                    predicate_pushdown_state.time_range_constraint, granularity, count
                )
            )
            logger.debug(
                LazyFormat(lambda: f"Adjusted time range constraint to: {cumulative_metric_adjusted_time_constraint}")
            )

        required_linkable_specs = self.__get_required_linkable_specs(
            queried_linkable_specs=queried_linkable_specs,
            filter_specs=metric_input_measure_spec.filter_spec_set.all_filter_specs,
            measure_spec_properties=measure_properties,
        )

        before_aggregation_time_spine_join_description = (
            metric_input_measure_spec.before_aggregation_time_spine_join_description
        )

        if measure_recipe is None:
            logger.debug(
                LazyFormat(
                    lambda: "Looking for a recipe to get:"
                    + indent(f"\nmeasure_specs:\n{mf_pformat([measure_spec])}")
                    + indent(f"\nevaluation:\n{mf_pformat(required_linkable_specs)}")
                )
            )
            measure_time_constraint = (
                (cumulative_metric_adjusted_time_constraint or predicate_pushdown_state.time_range_constraint)
                # If joining to time spine for time offset, constraints will be applied after that join.
                if not before_aggregation_time_spine_join_description
                else None
            )
            if measure_time_constraint is None:
                measure_pushdown_state = PredicatePushdownState.without_time_range_constraint(predicate_pushdown_state)
            else:
                measure_pushdown_state = PredicatePushdownState.with_time_range_constraint(
                    predicate_pushdown_state, time_range_constraint=measure_time_constraint
                )

            find_recipe_start_time = time.time()
            measure_recipe = self._find_source_node_recipe(
                FindSourceNodeRecipeParameterSet(
                    measure_spec_properties=measure_properties,
                    predicate_pushdown_state=measure_pushdown_state,
                    linkable_spec_set=required_linkable_specs,
                )
            )
            logger.debug(
                LazyFormat(
                    lambda: f"With {len(self._source_node_set.source_nodes_for_metric_queries)} source nodes, finding a recipe "
                    f"took {time.time() - find_recipe_start_time:.2f}s"
                )
            )

        logger.debug(LazyFormat(lambda: f"Using recipe:\n{indent(mf_pformat(measure_recipe))}"))

        if measure_recipe is None:
            raise UnableToSatisfyQueryError(
                f"Unable to join all items in request. Measure: {measure_spec.element_name}; Specs to join: {required_linkable_specs}"
            )

        base_queried_agg_time_dimension_specs = self._get_base_agg_time_dimensions(
            queried_linkable_specs=queried_linkable_specs, measure_reference=measure_spec.reference
        )

        # If a cumulative metric is queried with metric_time / agg_time_dimension, join over time range.
        # Otherwise, the measure will be aggregated over all time.
        unaggregated_measure_node: DataflowPlanNode = measure_recipe.source_node
        if cumulative and base_queried_agg_time_dimension_specs:
            unaggregated_measure_node = JoinOverTimeRangeNode.create(
                parent_node=unaggregated_measure_node,
                queried_agg_time_dimension_specs=base_queried_agg_time_dimension_specs,
                window=cumulative_window,
                grain_to_date=cumulative_grain_to_date,
                # Note: we use the original constraint here because the JoinOverTimeRangeNode will eventually get
                # rendered with an interval that expands the join window
                time_range_constraint=(
                    predicate_pushdown_state.time_range_constraint
                    if not before_aggregation_time_spine_join_description
                    else None
                ),
            )

        # If querying an offset metric, join to time spine before aggregation.
        if before_aggregation_time_spine_join_description and base_queried_agg_time_dimension_specs:
            # TODO: move all of this to a helper function
            assert before_aggregation_time_spine_join_description.join_type is SqlJoinType.INNER, (
                f"Expected {SqlJoinType.INNER} for joining to time spine before aggregation. Remove this if there's a "
                f"new use case."
            )

            join_on_time_dimension_spec = self._determine_time_spine_join_spec(
                measure_properties=measure_properties, required_time_spine_specs=base_queried_agg_time_dimension_specs
            )
            required_time_spine_specs = base_queried_agg_time_dimension_specs
            if join_on_time_dimension_spec not in required_time_spine_specs:
                required_time_spine_specs = (join_on_time_dimension_spec,) + required_time_spine_specs
            time_spine_node = self._build_time_spine_node(required_time_spine_specs)
            unaggregated_measure_node = JoinToTimeSpineNode.create(
                metric_source_node=unaggregated_measure_node,
                time_spine_node=time_spine_node,
                requested_agg_time_dimension_specs=base_queried_agg_time_dimension_specs,
                join_on_time_dimension_spec=join_on_time_dimension_spec,
                offset_window=before_aggregation_time_spine_join_description.offset_window,
                offset_to_grain=before_aggregation_time_spine_join_description.offset_to_grain,
                join_type=before_aggregation_time_spine_join_description.join_type,
            )

        custom_granularity_specs_to_join = [
            spec
            for spec in required_linkable_specs.time_dimension_specs_with_custom_grain
            # If this is the second layer of aggregation for a conversion metric, we have already joined the custom granularity.
            if spec not in measure_recipe.all_linkable_specs_required_for_source_nodes.as_tuple
        ]
        # Apply original time constraint if it wasn't applied to the source node recipe. For cumulative metrics, the constraint
        # may have been expanded and needs to be narrowed here. For offsets, the constraint was deferred to after the offset.
        # TODO - Pushdown: Encapsulate all of this window sliding bookkeeping in the pushdown params object
        time_range_constraint_to_apply = None
        if cumulative_metric_adjusted_time_constraint or before_aggregation_time_spine_join_description:
            time_range_constraint_to_apply = predicate_pushdown_state.time_range_constraint
        unaggregated_measure_node = self._build_pre_aggregation_plan(
            source_node=unaggregated_measure_node,
            join_targets=measure_recipe.join_targets,
            custom_granularity_specs=custom_granularity_specs_to_join,
            where_filter_specs=metric_input_measure_spec.filter_spec_set.all_filter_specs,
            time_range_constraint=time_range_constraint_to_apply,
            filter_to_specs=InstanceSpecSet(measure_specs=(measure_spec,)).merge(
                InstanceSpecSet.create_from_specs(queried_linkable_specs.as_tuple)
            ),
            measure_properties=measure_properties,
            queried_linkable_specs_for_semi_additive_join=queried_linkable_specs,
        )

        aggregate_measures_node = AggregateMeasuresNode.create(
            parent_node=unaggregated_measure_node, metric_input_measure_specs=(metric_input_measure_spec,)
        )

        # Joining to time spine after aggregation is for measures that specify `join_to_timespine: true` in the YAML spec.
        after_aggregation_time_spine_join_description = (
            metric_input_measure_spec.after_aggregation_time_spine_join_description
        )
        queried_agg_time_dimension_specs = queried_linkable_specs.included_agg_time_dimension_specs_for_measure(
            measure_reference=measure_spec.reference, semantic_model_lookup=self._semantic_model_lookup
        )
        if after_aggregation_time_spine_join_description and queried_agg_time_dimension_specs:
            # TODO: move all of this to a helper function
            assert after_aggregation_time_spine_join_description.join_type is SqlJoinType.LEFT_OUTER, (
                f"Expected {SqlJoinType.LEFT_OUTER} for joining to time spine after aggregation. Remove this if "
                f"there's a new use case."
            )
            time_spine_required_specs = copy.deepcopy(queried_agg_time_dimension_specs)

            # Find filters that contain only metric_time or agg_time_dimension. They will be applied to the time spine table.
            agg_time_only_filters: List[WhereFilterSpec] = []
            non_agg_time_filters: List[WhereFilterSpec] = []
            for filter_spec in metric_input_measure_spec.filter_spec_set.after_measure_aggregation_filter_specs:
                included_agg_time_specs = filter_spec.linkable_spec_set.included_agg_time_dimension_specs_for_measure(
                    measure_reference=measure_spec.reference, semantic_model_lookup=self._semantic_model_lookup
                )
                if set(included_agg_time_specs) == set(filter_spec.linkable_spec_set.as_tuple):
                    agg_time_only_filters.append(filter_spec)
                    for agg_time_spec in included_agg_time_specs:
                        if agg_time_spec not in time_spine_required_specs:
                            time_spine_required_specs.append(agg_time_spec)
                else:
                    non_agg_time_filters.append(filter_spec)

            time_spine_node = self._build_time_spine_node(
                queried_time_spine_specs=queried_agg_time_dimension_specs,
                time_range_constraint=predicate_pushdown_state.time_range_constraint,
                where_filter_specs=agg_time_only_filters,
            )
            output_node: DataflowPlanNode = JoinToTimeSpineNode.create(
                metric_source_node=aggregate_measures_node,
                time_spine_node=time_spine_node,
                requested_agg_time_dimension_specs=queried_agg_time_dimension_specs,
                join_on_time_dimension_spec=self._sort_by_base_granularity(queried_agg_time_dimension_specs)[0],
                join_type=after_aggregation_time_spine_join_description.join_type,
            )

            # Since new rows might have been added due to time spine join, re-apply constraints here. Only re-apply filters
            # for specs that are also in the queried specs, since those are the only ones that could change after the time
            # spine join. Exclude filters that contain only metric_time or an agg time dimension, since were already applied
            # to the time spine table.
            queried_non_agg_time_filter_specs = [
                filter_spec
                for filter_spec in non_agg_time_filters
                if set(filter_spec.linkable_specs).issubset(set(queried_linkable_specs.as_tuple))
            ]
            if len(queried_non_agg_time_filter_specs) > 0:
                output_node = WhereConstraintNode.create(
                    parent_node=output_node, where_specs=queried_non_agg_time_filter_specs, always_apply=True
                )

            # TODO: this will break if you query by agg_time_dimension but apply a time constraint on metric_time.
            # To fix when enabling time range constraints for agg_time_dimension.
            if queried_agg_time_dimension_specs and predicate_pushdown_state.time_range_constraint is not None:
                output_node = ConstrainTimeRangeNode.create(
                    parent_node=output_node, time_range_constraint=predicate_pushdown_state.time_range_constraint
                )
            return output_node

        return aggregate_measures_node

    def _build_pre_aggregation_plan(
        self,
        source_node: DataflowPlanNode,
        filter_to_specs: Optional[InstanceSpecSet] = None,
        join_targets: List[JoinDescription] = [],
        custom_granularity_specs: Sequence[TimeDimensionSpec] = (),
        where_filter_specs: Sequence[WhereFilterSpec] = (),
        time_range_constraint: Optional[TimeRangeConstraint] = None,
        measure_properties: Optional[MeasureSpecProperties] = None,
        queried_linkable_specs_for_semi_additive_join: Optional[LinkableSpecSet] = None,
        distinct: bool = False,
    ) -> DataflowPlanNode:
        """Adds standard pre-aggegation steps after building source node and before aggregation."""
        output_node = source_node
        if join_targets:
            output_node = JoinOnEntitiesNode.create(left_node=output_node, join_targets=join_targets)

        for custom_granularity_spec in custom_granularity_specs:
            output_node = JoinToCustomGranularityNode.create(
                parent_node=output_node, time_dimension_spec=custom_granularity_spec
            )

        if len(where_filter_specs) > 0:
            output_node = WhereConstraintNode.create(parent_node=output_node, where_specs=where_filter_specs)

        if time_range_constraint:
            output_node = ConstrainTimeRangeNode.create(
                parent_node=output_node, time_range_constraint=time_range_constraint
            )

        if measure_properties and measure_properties.non_additive_dimension_spec:
            if queried_linkable_specs_for_semi_additive_join is None:
                raise ValueError(
                    "`queried_linkable_specs_for_semi_additive_join` must be provided in when building pre-aggregation plan "
                    "if `non_additive_dimension_spec` is present."
                )
            output_node = self._build_semi_additive_join_node(
                measure_properties=measure_properties,
                queried_linkable_specs=queried_linkable_specs_for_semi_additive_join,
                parent_node=output_node,
            )
        if filter_to_specs:
            output_node = FilterElementsNode.create(
                parent_node=output_node, include_specs=filter_to_specs, distinct=distinct
            )

        return output_node

    def _build_semi_additive_join_node(
        self,
        measure_properties: MeasureSpecProperties,
        queried_linkable_specs: LinkableSpecSet,
        parent_node: DataflowPlanNode,
    ) -> SemiAdditiveJoinNode:
        non_additive_dimension_spec = measure_properties.non_additive_dimension_spec
        assert (
            non_additive_dimension_spec
        ), "_build_semi_additive_join_node() should only be called if there is a non_additive_dimension_spec."
        # Apply semi additive join on the node
        agg_time_dimension = measure_properties.agg_time_dimension
        non_additive_dimension_grain = measure_properties.agg_time_dimension_grain
        queried_time_dimension_spec: Optional[TimeDimensionSpec] = self._find_non_additive_dimension_in_linkable_specs(
            agg_time_dimension=agg_time_dimension,
            linkable_specs=queried_linkable_specs.as_tuple,
            non_additive_dimension_spec=non_additive_dimension_spec,
        )
        time_dimension_spec = TimeDimensionSpec(
            # The NonAdditiveDimensionSpec name property is a plain element name
            element_name=non_additive_dimension_spec.name,
            entity_links=(),
            time_granularity=ExpandedTimeGranularity.from_time_granularity(non_additive_dimension_grain),
        )
        window_groupings = tuple(
            LinklessEntitySpec.from_element_name(name) for name in non_additive_dimension_spec.window_groupings
        )
        return SemiAdditiveJoinNode.create(
            parent_node=parent_node,
            entity_specs=window_groupings,
            time_dimension_spec=time_dimension_spec,
            agg_by_function=non_additive_dimension_spec.window_choice,
            queried_time_dimension_spec=queried_time_dimension_spec,
        )

    def _choose_time_spine_source(self, required_time_spine_specs: Sequence[TimeDimensionSpec]) -> TimeSpineSource:
        """Choose the time spine source that can satisfy the required time spine specs."""
        return TimeSpineSource.choose_time_spine_source(
            required_time_spine_specs=required_time_spine_specs,
            time_spine_sources=self._source_node_builder.time_spine_sources,
        )

    def _choose_time_spine_metric_time_node(
        self, required_time_spine_specs: Sequence[TimeDimensionSpec]
    ) -> MetricTimeDimensionTransformNode:
        """Return the MetricTimeDimensionTransform time spine node needed to satisfy the specs."""
        time_spine_source = self._choose_time_spine_source(required_time_spine_specs)
        return self._source_node_set.time_spine_metric_time_nodes[time_spine_source.base_granularity]

    def _choose_time_spine_read_node(self, time_spine_source: TimeSpineSource) -> ReadSqlSourceNode:
        """Return the MetricTimeDimensionTransform time spine node needed to satisfy the specs."""
        return self._source_node_set.time_spine_read_nodes[time_spine_source.base_granularity]

    def _build_time_spine_node(
        self,
        queried_time_spine_specs: Sequence[TimeDimensionSpec],
        where_filter_specs: Sequence[WhereFilterSpec] = (),
        time_range_constraint: Optional[TimeRangeConstraint] = None,
    ) -> DataflowPlanNode:
        """Return the time spine node needed to satisfy the specs."""
        required_time_spine_spec_set = self.__get_required_linkable_specs(
            queried_linkable_specs=LinkableSpecSet(time_dimension_specs=tuple(queried_time_spine_specs)),
            filter_specs=where_filter_specs,
        )
        required_time_spine_specs = required_time_spine_spec_set.time_dimension_specs

        # TODO: support multiple time spines here. Build node on the one with the smallest base grain.
        # Then, pass custom_granularity_specs into _build_pre_aggregation_plan if they aren't satisfied by smallest time spine.
        time_spine_source = self._choose_time_spine_source(required_time_spine_specs)
        read_node = self._choose_time_spine_read_node(time_spine_source)
        time_spine_data_set = self._node_data_set_resolver.get_output_data_set(read_node)

        # Change the column aliases to match the specs that were requested in the query.
        time_spine_node = AliasSpecsNode.create(
            parent_node=read_node,
            change_specs=tuple(
                SpecToAlias(
                    input_spec=time_spine_data_set.instance_from_time_dimension_grain_and_date_part(required_spec).spec,
                    output_spec=required_spec,
                )
                for required_spec in required_time_spine_specs
            ),
        )

        # If the base grain of the time spine isn't selected, it will have duplicate rows that need deduping.
        should_dedupe = ExpandedTimeGranularity.from_time_granularity(time_spine_source.base_granularity) not in {
            spec.time_granularity for spec in queried_time_spine_specs
        }

        return self._build_pre_aggregation_plan(
            source_node=time_spine_node,
            filter_to_specs=InstanceSpecSet(time_dimension_specs=tuple(queried_time_spine_specs)),
            time_range_constraint=time_range_constraint,
            where_filter_specs=where_filter_specs,
            distinct=should_dedupe,
        )

    def _sort_by_base_granularity(self, time_dimension_specs: Sequence[TimeDimensionSpec]) -> List[TimeDimensionSpec]:
        """Sort the time dimensions by their base granularity.

        Specs with date part will come after specs without it. Standard grains will come before custom.
        """
        return sorted(
            time_dimension_specs,
            key=lambda spec: (
                spec.date_part is not None,
                spec.time_granularity.is_custom_granularity,
                spec.time_granularity.base_granularity.to_int(),
            ),
        )

    def _determine_time_spine_join_spec(
        self, measure_properties: MeasureSpecProperties, required_time_spine_specs: Tuple[TimeDimensionSpec, ...]
    ) -> TimeDimensionSpec:
        """Determine the spec to join on for a time spine join.

        Defaults to metric_time if it is included in the request, else the agg_time_dimension.
        Will use the smallest available grain for the meeasure.
        """
        join_spec_grain = ExpandedTimeGranularity.from_time_granularity(measure_properties.agg_time_dimension_grain)
        join_on_time_dimension_spec = DataSet.metric_time_dimension_spec(time_granularity=join_spec_grain)
        if not LinkableSpecSet(time_dimension_specs=required_time_spine_specs).contains_metric_time:
            sample_agg_time_dimension_spec = required_time_spine_specs[0]
            join_on_time_dimension_spec = sample_agg_time_dimension_spec.with_grain_and_date_part(
                time_granularity=join_spec_grain, date_part=None
            )
        return join_on_time_dimension_spec



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Sequence, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass
class SpecToAlias:
    """A mapping of an input spec that should be aliased to match an output spec."""

    input_spec: InstanceSpec
    output_spec: InstanceSpec


@dataclass(frozen=True, eq=False)
class AliasSpecsNode(DataflowPlanNode, ABC):
    """Change the columns matching the key specs to match the value specs."""

    change_specs: Tuple[SpecToAlias, ...]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.change_specs) > 0, "Must have at least one value in change_specs for AliasSpecsNode."

    @staticmethod
    def create(parent_node: DataflowPlanNode, change_specs: Tuple[SpecToAlias, ...]) -> AliasSpecsNode:  # noqa: D102
        return AliasSpecsNode(parent_nodes=(parent_node,), change_specs=change_specs)

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_ALIAS_SPECS_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_alias_specs_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Change Column Aliases"""

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("change_specs", self.change_specs),)

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__) and other_node.change_specs == self.change_specs

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> AliasSpecsNode:  # noqa: D102
        assert len(new_parent_nodes) == 1, "AliasSpecsNode accepts exactly one parent node."
        return AliasSpecsNode.create(
            parent_node=new_parent_nodes[0],
            change_specs=self.change_specs,
        )



from __future__ import annotations

from enum import Enum

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType


def is_expansive(agg_type: AggregationType) -> bool:
    """Expansive ≝ Op( X ∪ Y ∪ ...) = Op( Op(X) ∪ Op(Y) ∪ ...).

    NOTE: COUNT is only expansive because it's transformed into a SUM agg during model transformation
    """
    return agg_type in (
        AggregationType.SUM,
        AggregationType.MIN,
        AggregationType.MAX,
        AggregationType.SUM_BOOLEAN,
        AggregationType.COUNT,
    )


def is_additive(agg_type: AggregationType) -> bool:
    """Indicates that if you sum values over a dimension grouping, you will still get an accurate result for this metric."""
    if agg_type is AggregationType.SUM or agg_type is AggregationType.SUM_BOOLEAN or agg_type is AggregationType.COUNT:
        return True
    elif (
        agg_type is AggregationType.MIN
        or agg_type is AggregationType.MAX
        or agg_type is AggregationType.COUNT_DISTINCT
        or agg_type is AggregationType.AVERAGE
        or agg_type is AggregationType.PERCENTILE
        or agg_type is AggregationType.MEDIAN
    ):
        return False
    else:
        assert_values_exhausted(agg_type)


def fill_nulls_with_0(agg_type: AggregationType) -> bool:
    """Indicates if charts should show 0 instead of null where there are gaps in data."""
    return agg_type in (
        AggregationType.SUM,
        AggregationType.COUNT_DISTINCT,
        AggregationType.SUM_BOOLEAN,
        AggregationType.COUNT,
    )


def can_limit_dimension_values(agg_type: AggregationType) -> bool:
    """Indicates if we can limit dimension values in charts.

    Currently, this means:
    1. The dimensions we care about most are the ones with the highest numeric values
    2. We can calculate the "other" column in the postprocessor (meaning the metric is expansive)
    """
    return agg_type in (AggregationType.SUM, AggregationType.SUM_BOOLEAN, AggregationType.COUNT)


class AggregationState(Enum):
    """Represents how the measure is aggregated."""

    # When reading from the source, the measure is considered non-aggregated.
    NON_AGGREGATED = "NON_AGGREGATED"
    PARTIAL = "PARTIAL"
    COMPLETE = "COMPLETE"

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}.{self.name}"



from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Sequence, Tuple

from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.partition_spec_set import PartitionSpecSet
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec

from metricflow.dataset.dataset_classes import DataSet


@dataclass(frozen=True)
class PartitionDimensionJoinDescription:
    """Describes which partition dimensions should be joined between nodes."""

    start_node_dimension_spec: DimensionSpec
    node_to_join_dimension_spec: DimensionSpec


@dataclass(frozen=True)
class PartitionTimeDimensionJoinDescription:
    """Describes which partition time dimensions should be joined between nodes."""

    start_node_time_dimension_spec: TimeDimensionSpec
    node_to_join_time_dimension_spec: TimeDimensionSpec


class PartitionJoinResolver:
    """When joining data sets, this class helps to figure out the necessary partition specs to join on."""

    def __init__(self, semantic_model_lookup: SemanticModelLookup) -> None:  # noqa: D107
        self._semantic_model_lookup = semantic_model_lookup

    def _get_partitions(self, spec_set: InstanceSpecSet) -> PartitionSpecSet:
        """Returns the specs from the instance set that correspond to partition specs."""
        partition_dimension_specs = tuple(
            dimension_spec
            for dimension_spec in spec_set.dimension_specs
            if self._semantic_model_lookup.dimension_lookup.get_invariant(dimension_spec.reference).is_partition
        )
        partition_time_dimension_specs = tuple(
            time_dimension_spec
            for time_dimension_spec in spec_set.time_dimension_specs
            if time_dimension_spec.reference != DataSet.metric_time_dimension_reference()
            and self._semantic_model_lookup.dimension_lookup.get_invariant(time_dimension_spec.reference).is_partition
        )

        return PartitionSpecSet(
            dimension_specs=partition_dimension_specs,
            time_dimension_specs=partition_time_dimension_specs,
        )

    @staticmethod
    def _get_simplest_dimension_spec(dimension_specs: Sequence[DimensionSpec]) -> DimensionSpec:
        """Return the time dimension spec with the fewest entity links."""
        assert len(dimension_specs) > 0
        sorted_dimension_specs = sorted(dimension_specs, key=lambda x: len(x.entity_links))
        return sorted_dimension_specs[0]

    def resolve_partition_dimension_joins(
        self, left_node_spec_set: InstanceSpecSet, node_to_join_spec_set: InstanceSpecSet
    ) -> Tuple[PartitionDimensionJoinDescription, ...]:
        """Figures out which partition dimensions to join on."""
        start_node_partitions = self._get_partitions(left_node_spec_set)
        join_node_partitions = self._get_partitions(node_to_join_spec_set)

        partition_join_descriptions = []

        start_node_dimension_element_names = tuple(
            OrderedDict.fromkeys(x.element_name for x in start_node_partitions.dimension_specs)
        )

        for element_name in start_node_dimension_element_names:
            start_node_spec = PartitionJoinResolver._get_simplest_dimension_spec(
                tuple(x for x in start_node_partitions.dimension_specs if x.element_name == element_name)
            )
            join_node_time_dimension_element_names = tuple(
                OrderedDict.fromkeys(x.element_name for x in join_node_partitions.dimension_specs)
            )
            if element_name in join_node_time_dimension_element_names:
                join_node_spec = PartitionJoinResolver._get_simplest_dimension_spec(
                    tuple(x for x in join_node_partitions.dimension_specs if x.element_name == element_name)
                )
                partition_join_descriptions.append(
                    PartitionDimensionJoinDescription(
                        start_node_dimension_spec=start_node_spec,
                        node_to_join_dimension_spec=join_node_spec,
                    )
                )

        return tuple(partition_join_descriptions)

    @staticmethod
    def _get_simplest_time_dimension_spec(time_dimension_specs: Sequence[TimeDimensionSpec]) -> TimeDimensionSpec:
        """Return the time dimension spec with the smallest granularity, then fewest entity links."""
        assert len(time_dimension_specs) > 0
        # TODO: [custom granularity] restructure this method to operate on partition collections instead, and
        # move this enforcement to the PartitionSpecSet
        assert all(not x.time_granularity.is_custom_granularity for x in time_dimension_specs), (
            f"Found custom granularity in partition time dimension specs {time_dimension_specs}, but time partitions "
            "can only use standard granularities as they are based on engine date/time types!"
        )

        sorted_specs = sorted(
            time_dimension_specs, key=lambda x: (x.time_granularity.base_granularity, len(x.entity_links))
        )
        return sorted_specs[0]

    def resolve_partition_time_dimension_joins(
        self, left_node_spec_set: InstanceSpecSet, node_to_join_spec_set: InstanceSpecSet
    ) -> Tuple[PartitionTimeDimensionJoinDescription, ...]:
        """Figures out which partition time dimensions to join on."""
        start_node_partitions = self._get_partitions(left_node_spec_set)
        join_node_partitions = self._get_partitions(node_to_join_spec_set)
        partition_join_descriptions: List[PartitionTimeDimensionJoinDescription] = []

        # Get all unique element names
        time_dimension_element_names = tuple(
            OrderedDict.fromkeys(x.element_name for x in start_node_partitions.time_dimension_specs)
        )
        for element_name in time_dimension_element_names:
            start_node_spec = PartitionJoinResolver._get_simplest_time_dimension_spec(
                tuple(x for x in start_node_partitions.time_dimension_specs if x.element_name == element_name)
            )
            join_node_time_dimension_element_names = tuple(
                OrderedDict.fromkeys(x.element_name for x in join_node_partitions.time_dimension_specs)
            )
            if element_name in join_node_time_dimension_element_names:
                join_node_spec = PartitionJoinResolver._get_simplest_time_dimension_spec(
                    tuple(x for x in join_node_partitions.time_dimension_specs if x.element_name == element_name)
                )
                partition_join_descriptions.append(
                    PartitionTimeDimensionJoinDescription(
                        start_node_time_dimension_spec=start_node_spec,
                        node_to_join_time_dimension_spec=join_node_spec,
                    )
                )
        return tuple(partition_join_descriptions)



from __future__ import annotations

from typing import Union

from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId, MetricFlowDag
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import GroupByItemResolutionNode
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    MetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)

ResolutionDagSinkNode = Union[QueryGroupByItemResolutionNode, MetricGroupByItemResolutionNode]


class GroupByItemResolutionDag(MetricFlowDag[GroupByItemResolutionNode]):
    """A DAG that represents how valid group-by-items are resolved.

    In this representation, group-by-item candidates flow from the root / source nodes to the sink / leaf node. The
    source nodes represent the measures, and the sink node is usually the node that represents the metric query.

    The candidates that reach the sink node should be the valid ones. Generally, the nodes intersect the candidates
    from the parents and pass the intersection down to the child node. If the intersection produces an empty set, an
    issue can be produced with appropriate context and passed down to the child node. This allows generation of specific
    reasons for why a group-by item isn't valid for a given configuration. The nodes can also filter the candidates to
    realize limitations appropriate to that node.
    """

    def __init__(self, sink_node: ResolutionDagSinkNode) -> None:  # noqa: D107
        super().__init__(
            dag_id=DagId.from_id_prefix(StaticIdPrefix.GROUP_BY_ITEM_RESOLUTION_DAG),
            sink_nodes=[sink_node],
        )
        self._sink_node = sink_node

    @property
    def sink_node(self) -> ResolutionDagSinkNode:  # noqa: D102
        return self._sink_node



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class StringInputParsingIssue(MetricFlowQueryResolutionIssue):
    """Describe an issue with the query where the input string doesn't match one of the known naming schemes."""

    input_str: str

    @staticmethod
    def from_parameters(input_str: str) -> StringInputParsingIssue:  # noqa: D102
        return StringInputParsingIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=MetricFlowQueryResolutionPath.empty_instance(),
            input_str=input_str,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return f"The input {repr(self.input_str)} does not match any of the known formats."

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> StringInputParsingIssue:
        return StringInputParsingIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            input_str=self.input_str,
        )



from __future__ import annotations

from dbt_semantic_interfaces.protocols.dimension import Dimension
from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.measure import Measure
from dbt_semantic_interfaces.protocols.metric import Metric
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    MeasureReference,
    MetricReference,
    SemanticModelReference,
)


class ObjectToReference:
    """Converting model objects to a model reference.

    With moving all the model objects to dbt-semantic-interfaces and having *Reference objects
    remain in MetricFlow, all the `Object.reference` properties should be removed from the model
    objects. This means, we would need to replace all usages of `Object.reference` to
    instantiating the *Reference object directly. This class organizes and keeps note of
    where those use cases were.
    """

    @staticmethod
    def from_semantic_model(semantic_model: SemanticModel) -> SemanticModelReference:  # noqa: D102
        return SemanticModelReference(semantic_model_name=semantic_model.name)

    @staticmethod
    def from_measure(measure: Measure) -> MeasureReference:  # noqa: D102
        return MeasureReference(element_name=measure.name)

    @staticmethod
    def from_entity(entity: Entity) -> EntityReference:  # noqa: D102
        return EntityReference(element_name=entity.name)

    @staticmethod
    def from_metric(metric: Metric) -> MetricReference:  # noqa: D102
        return MetricReference(element_name=metric.name)

    @staticmethod
    def from_dimension(dimension: Dimension) -> DimensionReference:  # noqa: D102
        return DimensionReference(element_name=dimension.name)



from __future__ import annotations

import random
import string


def random_id() -> str:
    """Generates an 8-digit random alphanumeric string."""
    alphabet = string.ascii_lowercase + string.digits
    # Characters that go below the line are visually unappealing, so don't use those.
    filtered_alphabet = [x for x in alphabet if x not in "gjpqy"]
    return "".join(random.choices(filtered_alphabet, k=8))



from __future__ import annotations

from dbt_semantic_interfaces.implementations.semantic_manifest import (
    PydanticSemanticManifest,
)
from dbt_semantic_interfaces.protocols import ProtocolHint
from dbt_semantic_interfaces.transformations.transform_rule import (
    SemanticManifestTransformRule,
)
from typing_extensions import override


class DedupeMetricInputMeasuresRule(ProtocolHint[SemanticManifestTransformRule[PydanticSemanticManifest]]):
    """Dedupe the input measures within a metric.

    This can be removed once the fix is in the dbt-core transformation.
    """

    @override
    def _implements_protocol(self) -> SemanticManifestTransformRule[PydanticSemanticManifest]:
        return self

    @staticmethod
    def transform_model(semantic_manifest: PydanticSemanticManifest) -> PydanticSemanticManifest:  # noqa: D102
        for metric in semantic_manifest.metrics:
            metric.type_params.input_measures = list(dict.fromkeys(metric.input_measures).keys())
        return semantic_manifest



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, Optional, Sequence

from dbt_semantic_interfaces.implementations.time_spine import PydanticTimeSpineCustomGranularityColumn
from dbt_semantic_interfaces.protocols import SemanticManifest
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.collection_helpers.lru_cache import typed_lru_cache
from metricflow_semantics.specs.time_dimension_spec import DEFAULT_TIME_GRANULARITY, TimeDimensionSpec
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class TimeSpineSource:
    """A calendar table. Should contain at least one column with dates/times that map to a standard granularity.

    Dates should be contiguous. May also contain custom granularity columns.
    """

    schema_name: Optional[str]
    table_name: str = "mf_time_spine"
    # Name of the column in the table that contains date/time values that map to a standard granularity.
    base_column: str = "ds"
    # The time granularity of the base column.
    base_granularity: TimeGranularity = DEFAULT_TIME_GRANULARITY
    db_name: Optional[str] = None
    custom_granularities: Sequence[PydanticTimeSpineCustomGranularityColumn] = ()

    @property
    def spine_table(self) -> SqlTable:
        """Table containing all dates."""
        return SqlTable(schema_name=self.schema_name, table_name=self.table_name, db_name=self.db_name)

    @staticmethod
    def build_standard_time_spine_sources(
        semantic_manifest: SemanticManifest,
    ) -> Dict[TimeGranularity, TimeSpineSource]:
        """Creates a time spine source based on what's in the manifest."""
        time_spine_sources = {
            time_spine.primary_column.time_granularity: TimeSpineSource(
                schema_name=time_spine.node_relation.schema_name,
                table_name=time_spine.node_relation.alias,
                db_name=time_spine.node_relation.database,
                base_column=time_spine.primary_column.name,
                base_granularity=time_spine.primary_column.time_granularity,
                custom_granularities=tuple(
                    [
                        PydanticTimeSpineCustomGranularityColumn(
                            name=custom_granularity.name, column_name=custom_granularity.column_name
                        )
                        for custom_granularity in time_spine.custom_granularities
                    ]
                ),
            )
            for time_spine in semantic_manifest.project_configuration.time_spines
        }

        # For backward compatibility: if legacy time spine config exists in the manifest, add that time spine here for
        # backward compatibility. Ignore it if there is a new time spine config with the same granularity.
        legacy_time_spines = semantic_manifest.project_configuration.time_spine_table_configurations
        for legacy_time_spine in legacy_time_spines:
            if legacy_time_spine.grain not in time_spine_sources:
                time_spine_table = SqlTable.from_string(legacy_time_spine.location)
                time_spine_sources[legacy_time_spine.grain] = TimeSpineSource(
                    schema_name=time_spine_table.schema_name,
                    table_name=time_spine_table.table_name,
                    db_name=time_spine_table.db_name,
                    base_column=legacy_time_spine.column_name,
                    base_granularity=legacy_time_spine.grain,
                )

        return time_spine_sources

    @staticmethod
    @typed_lru_cache
    def build_custom_time_spine_sources(time_spine_sources: Sequence[TimeSpineSource]) -> Dict[str, TimeSpineSource]:
        """Creates a set of time spine sources with custom granularities based on what's in the manifest."""
        return {
            custom_granularity.name: time_spine_source
            for time_spine_source in time_spine_sources
            for custom_granularity in time_spine_source.custom_granularities
        }

    @staticmethod
    def build_custom_granularities(time_spine_sources: Sequence[TimeSpineSource]) -> Dict[str, ExpandedTimeGranularity]:
        """Creates a set of supported custom granularities based on what's in the manifest."""
        return {
            custom_granularity.name: ExpandedTimeGranularity(
                name=custom_granularity.name, base_granularity=time_spine_source.base_granularity
            )
            for time_spine_source in time_spine_sources
            for custom_granularity in time_spine_source.custom_granularities
        }

    @staticmethod
    def choose_time_spine_source(
        required_time_spine_specs: Sequence[TimeDimensionSpec],
        time_spine_sources: Dict[TimeGranularity, TimeSpineSource],
    ) -> TimeSpineSource:
        """Determine which time spine sources to use to satisfy the given specs.

        Custom grains can only use the time spine where they are defined. For standard grains, this will choose the time
        spine with the largest granularity that is compatible with all required standard grains. This ensures max efficiency
        for the query by minimizing the number of time spine joins and the amount of aggregation required. Example:
        - Time spines available: SECOND, MINUTE, DAY
        - Time granularities needed for request: HOUR, DAY
        --> Selected time spine: MINUTE
        """
        assert required_time_spine_specs, (
            "Choosing time spine source requires time spine specs, but the `required_time_spine_specs` param is empty. "
            "This indicates internal misconfiguration."
        )

        # Each custom grain can only be satisfied by one time spine.
        custom_time_spines = TimeSpineSource.build_custom_time_spine_sources(tuple(time_spine_sources.values()))
        required_time_spines = {
            custom_time_spines[spec.time_granularity.name]
            for spec in required_time_spine_specs
            if spec.time_granularity.is_custom_granularity
        }

        # Standard grains can be satisfied by any time spine with a base grain that's <= the standard grain.
        smallest_required_standard_grain = min(
            spec.time_granularity.base_granularity for spec in required_time_spine_specs
        )
        compatible_time_spines_for_standard_grains = {
            grain: time_spine_source
            for grain, time_spine_source in time_spine_sources.items()
            if grain.to_int() <= smallest_required_standard_grain.to_int()
        }
        if len(compatible_time_spines_for_standard_grains) == 0:
            raise RuntimeError(
                f"This query requires a time spine with granularity {smallest_required_standard_grain.name} or smaller, which is not configured. "
                f"The smallest available time spine granularity is {min(time_spine_sources).name}, which is too large."
                "See documentation for how to configure a new time spine: https://docs.getdbt.com/docs/build/metricflow-time-spine"
            )

        # If the standard grains can't be satisfied by the same time spines as the custom grains, add the largest compatible one.
        if not required_time_spines.intersection(set(compatible_time_spines_for_standard_grains.values())):
            required_time_spines.add(time_spine_sources[max(compatible_time_spines_for_standard_grains)])

        if len(required_time_spines) != 1:
            raise RuntimeError(
                "Multiple time spines are required to satisfy the specs, but only one is supported per query currently. "
                f"Multiple will be supported in the future. Time spines required: {required_time_spines}."
            )

        return required_time_spines.pop()

    @property
    def data_set_description(self) -> str:
        """Description to be displayed when this time spine is used in a data set."""
        return f"Read From Time Spine '{self.table_name}'"



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Sequence

from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.sequential_id import SequentialIdGenerator
from metricflow_semantics.mf_logging.runtime import log_block_runtime
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from typing_extensions import override

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.plan_conversion.dataflow_to_sql import (
    DataflowNodeToSqlSubqueryVisitor,
)

if TYPE_CHECKING:
    from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup


class DataflowPlanNodeOutputDataSetResolver:
    """Given a node in a dataflow plan, figure out what is the data set output by that node.

    Recall that in the dataflow plan, the nodes represent computation, and the inputs and outputs of the nodes are
    data sets containing measures, dimensions, etc.

    Note: the term "dimension" is used below, but it actually refers to any LinkableInstance.

    This class is useful when computing the joins required to retrieve the dimensions associated with a measure. The
    object that figures out the joins (NodeEvaluatorForLinkableInstances) is given a set of dataflow plan nodes that
    output different dimensions (this information is not necessarily in the node itself).

    NodeEvaluatorForLinkableInstances needs the data set associated with a node so that it knows what dimensions and
    entities are in the node. This information is used to figure out whether that node is useful to join to in order
    to retrieve a particular dimension.

    In the simple case, if the input nodes are all ReadSqlSourceNodes, then the data set is a member variable of the
    node, and there would be no need to traverse the dataflow node hierarchy with this class. However, if it's not a
    single node, that information is not available as the dataflow plan nodes contain little metadata. In that case,
    this class is needed to traverse the dataflow plan and figure that out.

    The reason why the input nodes to NodeEvaluatorForLinkableInstances wouldn't always be all ReadSqlSourceNodes is to
    realize an easy way to handle multi-hop joins without making changes to NodeEvaluatorForLinkableInstances. If we
    generate a set of nodes that already include the multi-hop dimensions available, the join resolution logic becomes
    much simpler. For example, a node like:

    <JoinOnEntitiesNode>
        <!-- Join dim_users and dim_devices by device_id -->
        <ReadSqlSourceNode>
          <!-- Read from dim_users to get user_id, device_id -->
        <ReadSqlSourceNodes>
          <!-- Read from dim_devices device_id, platform -->
    </JoinOnEntitiesNode>

    would have the dimension user_id__device_id__platform available, so to NodeEvaluatorForLinkableInstances,
    it's the same problem as doing a single-hop join. This simplifies the join resolution logic, though now the input
    to NodeEvaluatorForLinkableInstances needs to contain nodes that include all possible multi-hop joins.

    The logic to figure the dataset output by a node is the same as DataflowToSqlQueryPlanConverter because the same
    information is needed for generating SQL queries, so inheriting from that. We may want to look later at making
    another class to have better separation of concerns.
    """

    def __init__(  # noqa: D107
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_manifest_lookup: SemanticManifestLookup,
        _node_to_output_data_set: Optional[Dict[DataflowPlanNode, SqlDataSet]] = None,
    ) -> None:
        self._node_to_output_data_set: Dict[DataflowPlanNode, SqlDataSet] = _node_to_output_data_set or {}
        self._column_association_resolver = column_association_resolver
        self._semantic_manifest_lookup = semantic_manifest_lookup
        self._to_data_set_visitor = _NodeDataSetVisitor(
            column_association_resolver=self._column_association_resolver,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )

    def get_output_data_set(self, node: DataflowPlanNode) -> SqlDataSet:
        """Cached since this will be called repeatedly during the computation of multiple metrics.

        # TODO: The cache needs to be pruned, but has not yet been an issue.
        """
        if node not in self._node_to_output_data_set:
            self._node_to_output_data_set[node] = node.accept(self._to_data_set_visitor)

        return self._node_to_output_data_set[node]

    def cache_output_data_sets(self, nodes: Sequence[DataflowPlanNode]) -> None:
        """Cache the output of the given nodes for consistent retrieval with `get_output_data_set`."""
        with log_block_runtime(f"cache_output_data_sets for {len(nodes)} nodes"):
            for node in nodes:
                self.get_output_data_set(node)

    def copy(self) -> DataflowPlanNodeOutputDataSetResolver:
        """Return a copy of this with the same nodes cached."""
        return DataflowPlanNodeOutputDataSetResolver(
            column_association_resolver=self._column_association_resolver,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
            _node_to_output_data_set=dict(self._node_to_output_data_set),
        )


class _NodeDataSetVisitor(DataflowNodeToSqlSubqueryVisitor):
    @override
    def _next_unique_table_alias(self) -> str:
        """Return the next unique table alias to use in generating queries.

        This uses a separate prefix in order to minimize subquery ID thrash.
        """
        return SequentialIdGenerator.create_next_id(StaticIdPrefix.NODE_RESOLVER_SUB_QUERY).str_value



from __future__ import annotations

from dbt_semantic_interfaces.references import TimeDimensionReference
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from dbt_semantic_interfaces.validations.unique_valid_name import MetricFlowReservedKeywords

from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

# Shortcuts for referring to the metric time dimension.
MTD = MetricFlowReservedKeywords.METRIC_TIME.value
MTD_REFERENCE = TimeDimensionReference(element_name=MetricFlowReservedKeywords.METRIC_TIME.value)
MTD_SPEC_DAY = TimeDimensionSpec(
    element_name=MetricFlowReservedKeywords.METRIC_TIME.value,
    entity_links=(),
    time_granularity=ExpandedTimeGranularity.from_time_granularity(TimeGranularity.DAY),
)
MTD_SPEC_WEEK = TimeDimensionSpec(
    element_name=MetricFlowReservedKeywords.METRIC_TIME.value,
    entity_links=(),
    time_granularity=ExpandedTimeGranularity.from_time_granularity(TimeGranularity.WEEK),
)
MTD_SPEC_MONTH = TimeDimensionSpec(
    element_name=MetricFlowReservedKeywords.METRIC_TIME.value,
    entity_links=(),
    time_granularity=ExpandedTimeGranularity.from_time_granularity(TimeGranularity.MONTH),
)
MTD_SPEC_QUARTER = TimeDimensionSpec(
    element_name=MetricFlowReservedKeywords.METRIC_TIME.value,
    entity_links=(),
    time_granularity=ExpandedTimeGranularity.from_time_granularity(TimeGranularity.QUARTER),
)
MTD_SPEC_YEAR = TimeDimensionSpec(
    element_name=MetricFlowReservedKeywords.METRIC_TIME.value,
    entity_links=(),
    time_granularity=ExpandedTimeGranularity.from_time_granularity(TimeGranularity.YEAR),
)



from __future__ import annotations

from dbt_semantic_interfaces.references import SemanticModelReference
from metricflow_semantics.instances import InstanceSet
from typing_extensions import override

from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.sql.sql_plan import SqlSelectStatementNode


class SemanticModelDataSet(SqlDataSet):
    """Similar to SqlDataSet, but contains metadata on the semantic model that was used to create this."""

    def __init__(  # noqa: D107
        self,
        semantic_model_reference: SemanticModelReference,
        instance_set: InstanceSet,
        sql_select_node: SqlSelectStatementNode,
    ) -> None:
        self._semantic_model_reference = semantic_model_reference
        super().__init__(instance_set=instance_set, sql_select_node=sql_select_node)

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}({repr(self._semantic_model_reference.semantic_model_name)})"

    @property
    @override
    def semantic_model_reference(self) -> SemanticModelReference:
        return self._semantic_model_reference



from __future__ import annotations

from typing import Collection

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from metricflow_semantics.errors.error_classes import UnsupportedEngineFeatureError
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_exprs import (
    SqlGenerateUuidExpression,
    SqlPercentileExpression,
    SqlPercentileFunctionType,
)
from typing_extensions import override

from metricflow.protocols.sql_client import SqlEngine
from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlPlanRenderer


class SnowflakeSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the Snowflake engine."""

    sql_engine = SqlEngine.SNOWFLAKE

    @property
    @override
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        return {
            SqlPercentileFunctionType.CONTINUOUS,
            SqlPercentileFunctionType.DISCRETE,
            SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS,
        }

    @override
    def render_date_part(self, date_part: DatePart) -> str:
        if date_part is DatePart.DOW:
            return "dayofweekiso"

        return super().render_date_part(date_part)

    @override
    def visit_generate_uuid_expr(self, node: SqlGenerateUuidExpression) -> SqlExpressionRenderResult:
        return SqlExpressionRenderResult(
            sql="UUID_STRING()",
            bind_parameter_set=SqlBindParameterSet(),
        )

    @override
    def visit_percentile_expr(self, node: SqlPercentileExpression) -> SqlExpressionRenderResult:
        """Render a percentile expression for Snowflake."""
        arg_rendered = self.render_sql_expr(node.order_by_arg)
        params = arg_rendered.bind_parameter_set
        percentile = node.percentile_args.percentile

        if node.percentile_args.function_type is SqlPercentileFunctionType.CONTINUOUS:
            function_str = "PERCENTILE_CONT"
        elif node.percentile_args.function_type is SqlPercentileFunctionType.DISCRETE:
            function_str = "PERCENTILE_DISC"
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS:
            return SqlExpressionRenderResult(
                sql=f"APPROX_PERCENTILE({arg_rendered.sql}, {percentile})",
                bind_parameter_set=params,
            )
        elif node.percentile_args.function_type is SqlPercentileFunctionType.APPROXIMATE_DISCRETE:
            raise UnsupportedEngineFeatureError(
                "Approximate discrete percentile aggregate not supported for Snowflake. Set "
                + "use_discrete_percentile and/or use_approximate_percentile to false in all percentile measures."
            )
        else:
            assert_values_exhausted(node.percentile_args.function_type)

        return SqlExpressionRenderResult(
            sql=f"{function_str}({percentile}) WITHIN GROUP (ORDER BY ({arg_rendered.sql}))",
            bind_parameter_set=params,
        )


class SnowflakeSqlPlanRenderer(DefaultSqlPlanRenderer):
    """Plan renderer for the Snowflake engine."""

    EXPR_RENDERER = SnowflakeSqlExpressionRenderer()

    @property
    @override
    def expr_renderer(self) -> SqlExpressionRenderer:
        return self.EXPR_RENDERER



from __future__ import annotations

import logging
import typing
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.references import MetricReference
from dbt_semantic_interfaces.type_enums import MetricType
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat, mf_pformat_dict
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.query.group_by_item.candidate_push_down.group_by_item_candidate import GroupByItemCandidateSet
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import (
    WhereFilterLocation,
    WhereFilterLocationType,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.measure_source_node import (
    MeasureGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.metric_resolution_node import (
    MetricGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.no_metrics_query_source_node import (
    NoMetricsGroupByItemSourceNode,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.query_resolution_node import (
    QueryGroupByItemResolutionNode,
)
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.group_by_item_resolver.invalid_use_of_date_part import (
    MetricExcludesDatePartIssue,
)
from metricflow_semantics.query.issues.group_by_item_resolver.no_common_items import NoCommonItemsInParents
from metricflow_semantics.query.issues.group_by_item_resolver.no_matching_items_for_measure import (
    NoMatchingItemsForMeasure,
)
from metricflow_semantics.query.issues.group_by_item_resolver.no_matching_items_for_no_metrics_query import (
    NoMatchingItemsForNoMetricsQuery,
)
from metricflow_semantics.query.issues.group_by_item_resolver.no_parent_candidates import NoParentCandidates
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryResolutionIssueSet,
)
from metricflow_semantics.query.suggestion_generator import QueryItemSuggestionGenerator
from metricflow_semantics.specs.patterns.none_date_part import NoneDatePartPattern
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern

if typing.TYPE_CHECKING:
    from metricflow_semantics.query.group_by_item.group_by_item_resolver import GroupByItemResolver


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PushDownResult:
    """The result that is pushed down from the root nodes to the leaf nodes to resolve ambiguous group-by items."""

    # The set of candidate specs that could match the ambiguous group-by item.
    candidate_set: GroupByItemCandidateSet
    # The issues seen so far while pushing down the result / resolving the ambiguity.
    issue_set: MetricFlowQueryResolutionIssueSet
    # The largest default time granularity of the metrics seen in the DAG so far. Used to resolve metric_time.
    # TODO: [custom granularity] decide whether or not to support custom granularities as metric_time defaults and
    # update accordingly
    max_metric_default_time_granularity: Optional[TimeGranularity] = None

    def __post_init__(self) -> None:  # noqa: D105
        # If there are no errors, there should be a candidate spec in each candidate set.
        # If there are errors, there shouldn't be any candidate sets.
        assert (not self.issue_set.has_errors and not self.candidate_set.is_empty) or (
            self.issue_set.has_errors and self.candidate_set.is_empty
        ), mf_pformat_dict(
            "candidate_set / issue_set mismatch:", {"candidate_set": self.candidate_set, "issue_set": self.issue_set}
        )

    def filter_candidates_by_pattern(self, spec_pattern: SpecPattern) -> PushDownResult:
        """Create a new result where only the candidates matching the pattern are included."""
        return PushDownResult(
            candidate_set=self.candidate_set.filter_candidates_by_pattern(spec_pattern),
            issue_set=self.issue_set,
            max_metric_default_time_granularity=self.max_metric_default_time_granularity,
        )

    def filter_candidates_by_patterns(self, spec_patterns: Sequence[SpecPattern]) -> PushDownResult:
        """Create a new result where only the candidates matching the patterns are included."""
        if len(spec_patterns) == 0:
            return self

        candidate_set = self.candidate_set
        for spec_pattern in spec_patterns:
            candidate_set = candidate_set.filter_candidates_by_pattern(spec_pattern)

        return PushDownResult(
            candidate_set=candidate_set,
            issue_set=self.issue_set,
            max_metric_default_time_granularity=self.max_metric_default_time_granularity,
        )


class DagTraversalPathTracker:
    """Helps track the path traveled by a visitor through the nodes in a group-by-tem resolution DAG."""

    def __init__(self) -> None:  # noqa: D107
        self._current_path: List[GroupByItemResolutionNode] = []

    @contextmanager
    def track_node_visit(self, node: GroupByItemResolutionNode) -> Iterator[MetricFlowQueryResolutionPath]:
        """Context manager used to wrap the visit calls.

        e.g.

            def visit_some_node(self, node) -> Result:
                with track_node_visit(node) as current_traversal_path:
                    # current_traversal_path is the path from the start node to this node.
                # context manager removes this node from current_traversal_path so sibling branches don't include it
        """
        self._current_path.append(node)
        yield MetricFlowQueryResolutionPath(tuple(self._current_path))
        self._current_path.pop(-1)


class _PushDownGroupByItemCandidatesVisitor(GroupByItemResolutionNodeVisitor[PushDownResult]):
    """A visitor that implements the logic for group-by-item resolution using the resolution DAG.

    Please see the behavior at each node for more details. Overall, this visitor pushes candidates that match an
    ambiguous group-by-item pattern from the root nodes (representing measures / the specs available for those
    measures) to the leaf node (the query for metrics / valid group-by items for querying). For nodes with multiple
    parents, the candidates from each parent are intersected and passed to the child node.
    """

    def __init__(
        self,
        manifest_lookup: SemanticManifestLookup,
        suggestion_generator: Optional[QueryItemSuggestionGenerator],
        group_by_item_resolver: GroupByItemResolver,
        source_spec_patterns: Sequence[SpecPattern] = (),
        filter_location: Optional[WhereFilterLocation] = None,
    ) -> None:
        """Initializer.

        Args:
            manifest_lookup: The semantic manifest lookup associated with the resolution DAG that this will traverse.
            group_by_item_resolver: The group-by-item resolver for the query to help generate suggestions.
            suggestion_generator: If there are issues with matching patterns to specs, use this to generate suggestions
            that will go in the issue.
            source_spec_patterns: The patterns to apply to the specs available at the measure nodes.
            LinkableElementProperty).
            filter_location: If resolving a where filter item, where this filter was defined.
        """
        self._semantic_manifest_lookup = manifest_lookup
        self._source_spec_patterns = tuple(source_spec_patterns)
        self._path_from_start_node_tracker = DagTraversalPathTracker()
        self._suggestion_generator = suggestion_generator
        self._filter_location = filter_location
        self._group_by_item_resolver_for_query = group_by_item_resolver

    @override
    def visit_measure_node(self, node: MeasureGroupByItemSourceNode) -> PushDownResult:
        """Push the group-by-item specs that are available to the measure and match the source patterns to the child."""
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            logger.debug(LazyFormat(lambda: f"Handling {node.ui_description}"))

            items_available_for_measure = self._semantic_manifest_lookup.metric_lookup.linkable_elements_for_measure(
                measure_reference=node.measure_reference,
                element_filter=LinkableElementFilter.merge_iterable(
                    spec_pattern.element_pre_filter for spec_pattern in self._source_spec_patterns
                ),
            )

            # The following is needed to handle limitation of cumulative metrics. Filtering could be done at the measure
            # node, but doing it here makes it a little easier to generate the error message.
            metric = self._semantic_manifest_lookup.metric_lookup.get_metric(node.child_metric_reference)

            patterns_to_apply: Tuple[SpecPattern, ...] = ()
            if metric.type is MetricType.SIMPLE or metric.type is MetricType.CONVERSION:
                pass
            elif metric.type is MetricType.RATIO or metric.type is MetricType.DERIVED:
                assert False, f"A measure should have a simple or cumulative metric as a child, but got {metric.type}"
            elif metric.type is MetricType.CUMULATIVE:
                patterns_to_apply = (
                    # Date part doesn't make clear sense with cumulative metrics, so we don't allow it.
                    NoneDatePartPattern(),
                )
            else:
                assert_values_exhausted(metric.type)

            matching_items = items_available_for_measure.filter_by_spec_patterns(
                patterns_to_apply + self._source_spec_patterns
            )

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    LazyFormat(
                        lambda: f"For {node.ui_description}:\n"
                        + indent(
                            "After applying patterns:\n"
                            + indent(mf_pformat(patterns_to_apply + self._source_spec_patterns))
                            + "\n"
                            + "to inputs, matches are:\n"
                            + indent(mf_pformat(matching_items.specs))
                        )
                    )
                )

            # The specified patterns don't match to any of the available group-by-items that can be queried for the
            # measure.
            if matching_items.spec_count == 0:
                input_suggestions: Sequence[str] = ()

                if self._suggestion_generator is not None:
                    candidate_specs = self._group_by_item_resolver_for_query.resolve_available_items(
                        source_spec_patterns=self._suggestion_generator.candidate_filters
                    ).specs
                    input_suggestions = self._suggestion_generator.input_suggestions(tuple(candidate_specs))

                return PushDownResult(
                    candidate_set=GroupByItemCandidateSet.empty_instance(),
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        NoMatchingItemsForMeasure.from_parameters(
                            parent_issues=(),
                            query_resolution_path=current_traversal_path,
                            input_suggestions=input_suggestions,
                        )
                    ),
                )

            return PushDownResult(
                candidate_set=GroupByItemCandidateSet(
                    measure_paths=(current_traversal_path,),
                    linkable_element_set=matching_items,
                    path_from_leaf_node=current_traversal_path,
                ),
                issue_set=MetricFlowQueryResolutionIssueSet(),
            )

    def _merge_push_down_results_from_parents(
        self,
        push_down_results_from_parents: Dict[GroupByItemResolutionNode, PushDownResult],
        current_traversal_path: MetricFlowQueryResolutionPath,
    ) -> PushDownResult:
        """Helper method that merges push down results from parents into a single result.

        This logic is shared between several nodes. In general, a node intersects the candidate sets from the parents
        to form a new candidate set, and passes that down to the child node. This handles generation of an issue if
        the parent candidate sets don't have a common candidate.
        """
        merged_issue_set: MetricFlowQueryResolutionIssueSet = MetricFlowQueryResolutionIssueSet.merge_iterable(
            parent_candidate_set.issue_set for parent_candidate_set in push_down_results_from_parents.values()
        )
        if len(push_down_results_from_parents) == 0:
            # If there are no results of any kind from the parents it means there are no parents, which means
            # something is likely wrong with the semantic manifest or the query construction itself
            merged_issue_set = merged_issue_set.add_issue(
                NoParentCandidates.from_parameters(query_resolution_path=current_traversal_path)
            )

        metric_default_time_granularities = {
            parent_candidate_set.max_metric_default_time_granularity
            for parent_candidate_set in push_down_results_from_parents.values()
            if parent_candidate_set.max_metric_default_time_granularity
        }
        max_metric_default_time_granularity = (
            max(metric_default_time_granularities) if metric_default_time_granularities else None
        )
        if merged_issue_set.has_errors:
            return PushDownResult(
                candidate_set=GroupByItemCandidateSet.empty_instance(),
                issue_set=merged_issue_set,
                max_metric_default_time_granularity=max_metric_default_time_granularity,
            )

        parent_candidate_sets = tuple(
            parent_candidate_set.candidate_set for parent_candidate_set in push_down_results_from_parents.values()
        )
        intersected_candidate_set = GroupByItemCandidateSet.intersection(
            path_from_leaf_node=current_traversal_path, candidate_sets=parent_candidate_sets
        )

        if intersected_candidate_set.is_empty:
            return PushDownResult(
                candidate_set=intersected_candidate_set,
                issue_set=merged_issue_set.add_issue(
                    NoCommonItemsInParents.from_parameters(
                        query_resolution_path=current_traversal_path,
                        parent_node_to_candidate_set={
                            parent_node: push_down_result.candidate_set
                            for parent_node, push_down_result in push_down_results_from_parents.items()
                        },
                        parent_issues=(),
                    )
                ),
                max_metric_default_time_granularity=max_metric_default_time_granularity,
            )
        return PushDownResult(
            candidate_set=intersected_candidate_set,
            issue_set=merged_issue_set,
            max_metric_default_time_granularity=max_metric_default_time_granularity,
        )

    @override
    def visit_metric_node(self, node: MetricGroupByItemResolutionNode) -> PushDownResult:
        """At the metric node, intersect candidates from the parents and pass them to the children.

        This node can represent a metric that prevents querying by some types of group-by items. To model that
        restriction, this filters the appropriate candidates.
        """
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            merged_result_from_parents = self._merge_push_down_results_from_parents(
                push_down_results_from_parents={
                    parent_node: parent_node.accept(self) for parent_node in node.parent_nodes
                },
                current_traversal_path=current_traversal_path,
            )
            logger.debug(LazyFormat(lambda: f"Handling {node.ui_description}"))
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    LazyFormat(
                        lambda: "Candidates from parents:\n"
                        + indent(mf_pformat(merged_result_from_parents.candidate_set.specs))
                    )
                )
            if merged_result_from_parents.candidate_set.is_empty:
                return merged_result_from_parents

            metric = self._semantic_manifest_lookup.metric_lookup.get_metric(node.metric_reference)

            # For metrics with offset_to_grain, don't allow date_part group-by-items
            patterns_to_apply: Sequence[SpecPattern] = ()
            if (
                metric.type is MetricType.SIMPLE
                or metric.type is MetricType.CUMULATIVE
                or metric.type is MetricType.CONVERSION
            ):
                pass
            elif metric.type is MetricType.RATIO or metric.type is MetricType.DERIVED:
                for input_metric in metric.input_metrics:
                    if input_metric.offset_to_grain:
                        # From comment in previous query parser:
                        # "Cannot extract date part for metrics with offset_to_grain."
                        patterns_to_apply = (NoneDatePartPattern(),)
                        break
            else:
                assert_values_exhausted(metric.type)

            candidate_items = merged_result_from_parents.candidate_set.linkable_element_set
            issue_sets_to_merge = [merged_result_from_parents.issue_set]

            matched_items = candidate_items.filter_by_spec_patterns(patterns_to_apply)

            if logger.isEnabledFor(logging.DEBUG):
                matched_specs = matched_items.specs
                logger.debug(
                    LazyFormat(
                        lambda: f"For {node.ui_description}:\n"
                        + indent(
                            "After applying patterns:\n"
                            + indent(mf_pformat(patterns_to_apply))
                            + "\n"
                            + "to inputs, outputs are:\n"
                            + indent(mf_pformat(matched_specs))
                        )
                    )
                )

            # There were candidates that were common from the ones passed from parents, but after applying the filters,
            # none of the candidates were valid.
            if matched_items.spec_count == 0:
                issue_sets_to_merge.append(
                    MetricFlowQueryResolutionIssueSet.from_issue(
                        MetricExcludesDatePartIssue.from_parameters(
                            query_resolution_path=current_traversal_path,
                            candidate_specs=candidate_items.specs,
                            parent_issues=(),
                        )
                    )
                )

            metric_to_use_for_time_granularity_resolution = metric
            # If this is resolving a filter defined on an input metric, use the outer metric's time_granularity.
            if (
                node.metric_input_location
                and self._filter_location
                and self._filter_location.location_type == WhereFilterLocationType.INPUT_METRIC
            ):
                metric_to_use_for_time_granularity_resolution = self._semantic_manifest_lookup.metric_lookup.get_metric(
                    node.metric_input_location.derived_metric_reference
                )

            # If time granularity is not set for the metric, defaults to DAY if available, else the smallest available granularity.
            # Note: ignores any granularity set on input metrics.
            metric_default_time_granularity = metric_to_use_for_time_granularity_resolution.time_granularity or max(
                TimeGranularity.DAY,
                self._semantic_manifest_lookup.metric_lookup.get_min_queryable_time_granularity(
                    MetricReference(metric_to_use_for_time_granularity_resolution.name)
                ),
            )

            if matched_items.spec_count == 0:
                return PushDownResult(
                    candidate_set=GroupByItemCandidateSet.empty_instance(),
                    issue_set=MetricFlowQueryResolutionIssueSet.merge_iterable(issue_sets_to_merge),
                    max_metric_default_time_granularity=metric_default_time_granularity,
                )

            return PushDownResult(
                candidate_set=GroupByItemCandidateSet(
                    linkable_element_set=matched_items,
                    measure_paths=merged_result_from_parents.candidate_set.measure_paths,
                    path_from_leaf_node=current_traversal_path,
                ),
                issue_set=MetricFlowQueryResolutionIssueSet.merge_iterable(issue_sets_to_merge),
                max_metric_default_time_granularity=metric_default_time_granularity,
            )

    @override
    def visit_query_node(self, node: QueryGroupByItemResolutionNode) -> PushDownResult:
        """Intersect candidates from the parent node and then return the result.

        The query node is a leaf node in the DAG, so the result from here is handled outside of this visitor.
        """
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            merged_result_from_parents = self._merge_push_down_results_from_parents(
                push_down_results_from_parents={
                    parent_node: parent_node.accept(self) for parent_node in node.parent_nodes
                },
                current_traversal_path=current_traversal_path,
            )
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    LazyFormat(
                        lambda: "Candidates from parents:\n"
                        + indent(mf_pformat(merged_result_from_parents.candidate_set.specs))
                    )
                )

            return merged_result_from_parents

    @override
    def visit_no_metrics_query_node(self, node: NoMetricsGroupByItemSourceNode) -> PushDownResult:
        """Pass the group-by items that can be queried without metrics to the child node."""
        with self._path_from_start_node_tracker.track_node_visit(node) as current_traversal_path:
            logger.debug(LazyFormat(lambda: f"Handling {node.ui_description}"))
            # This is a case for distinct dimension values from semantic models.
            candidate_elements = self._semantic_manifest_lookup.metric_lookup.linkable_elements_for_no_metrics_query()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(LazyFormat(lambda: f"Candidate elements are:\n{mf_pformat(candidate_elements)}"))
            candidates_after_filtering = candidate_elements.filter_by_spec_patterns(self._source_spec_patterns)

            if candidates_after_filtering.spec_count == 0:
                return PushDownResult(
                    candidate_set=GroupByItemCandidateSet.empty_instance(),
                    issue_set=MetricFlowQueryResolutionIssueSet.from_issue(
                        NoMatchingItemsForNoMetricsQuery.from_parameters(
                            parent_issues=(),
                            query_resolution_path=current_traversal_path,
                        )
                    ),
                )

            return PushDownResult(
                candidate_set=GroupByItemCandidateSet(
                    linkable_element_set=candidates_after_filtering,
                    measure_paths=(current_traversal_path,),
                    path_from_leaf_node=current_traversal_path,
                ),
                issue_set=MetricFlowQueryResolutionIssueSet.empty_instance(),
            )
